import sys
import base64
import json


def parse_parameter_value(base64_value):
    val = base64.b64decode(base64_value).decode("utf-8")

    try:
        return json.loads(val)
    except json.decoder.JSONDecodeError:
        # fallback to using the original value.
        return val


if __name__ == "__main__":
    base64_val = sys.argv[1]

    print(parse_parameter_value(base64_val))



#!/usr/bin/env python
"""Get git repository information for the package

Functions to retrieve git repository details like URL, branch name,
and commit SHA for Metaflow code provenance tracking.
"""

import os
import subprocess
from typing import Dict, List, Optional, Tuple, Union

# Cache for git information to avoid repeated subprocess calls
_git_info_cache = None

__all__ = ("get_repository_info",)


def _call_git(
    args: List[str], path=Union[str, os.PathLike]
) -> Tuple[Optional[str], Optional[int], bool]:
    """
    Call git with provided args.

    Returns
    -------
        tuple : Tuple containing
            (stdout, exitcode, failure) of the call
    """
    try:
        result = subprocess.run(
            ["git", *args],
            cwd=path,
            capture_output=True,
            text=True,
            check=False,
        )
        return result.stdout.strip(), result.returncode, False
    except (OSError, subprocess.SubprocessError):
        # Covers subprocess timeouts and other errors which would not lead to an exit code
        return None, None, True


def _get_repo_url(path: Union[str, os.PathLike]) -> Optional[str]:
    """Get the repository URL from git config"""
    stdout, returncode, _failed = _call_git(
        ["config", "--get", "remote.origin.url"], path
    )
    if returncode == 0:
        url = stdout
        # Convert SSH URLs to HTTPS for clickable links
        if url.startswith("git@"):
            parts = url.split(":", 1)
            if len(parts) == 2:
                domain = parts[0].replace("git@", "")
                repo_path = parts[1]
                url = f"https://{domain}/{repo_path}"
        return url
    return None


def _get_branch_name(path: Union[str, os.PathLike]) -> Optional[str]:
    """Get the current git branch name"""
    stdout, returncode, _failed = _call_git(["rev-parse", "--abbrev-ref", "HEAD"], path)
    return stdout if returncode == 0 else None


def _get_commit_sha(path: Union[str, os.PathLike]) -> Optional[str]:
    """Get the current git commit SHA"""
    stdout, returncode, _failed = _call_git(["rev-parse", "HEAD"], path)
    return stdout if returncode == 0 else None


def _is_in_git_repo(path: Union[str, os.PathLike]) -> bool:
    """Check if we're currently in a git repository"""
    stdout, returncode, _failed = _call_git(
        ["rev-parse", "--is-inside-work-tree"], path
    )
    return returncode == 0 and stdout == "true"


def _has_uncommitted_changes(path: Union[str, os.PathLike]) -> Optional[bool]:
    """Check if the git repository has uncommitted changes"""
    _stdout, returncode, failed = _call_git(
        ["diff-index", "--quiet", "HEAD", "--"], path
    )
    if failed:
        return None
    return returncode != 0


def get_repository_info(path: Union[str, os.PathLike]) -> Dict[str, Union[str, bool]]:
    """Get git repository information for a path

    Returns:
        dict: Dictionary containing:
            repo_url: Repository URL (converted to HTTPS if from SSH)
            branch_name: Current branch name
            commit_sha: Current commit SHA
            has_uncommitted_changes: Boolean indicating if there are uncommitted changes
    """
    global _git_info_cache

    if _git_info_cache is not None:
        return _git_info_cache

    _git_info_cache = {}
    if _is_in_git_repo(path):
        _git_info_cache = {
            "repo_url": _get_repo_url(path),
            "branch_name": _get_branch_name(path),
            "commit_sha": _get_commit_sha(path),
            "has_uncommitted_changes": _has_uncommitted_changes(path),
        }

    return _git_info_cache



import os
from contextlib import contextmanager
from typing import Callable, Generator, List, Optional, Tuple

from ..util import to_unicode, walk_without_cycles


def walk(
    root: str,
    exclude_hidden: bool = True,
    file_filter: Optional[Callable[[str], bool]] = None,
    exclude_tl_dirs: Optional[List[str]] = None,
) -> Generator[Tuple[str, str], None, None]:
    root = to_unicode(root)  # handle files/folder with non ascii chars
    prefixlen = len("%s/" % os.path.dirname(root))
    for (
        path,
        _,
        files,
    ) in walk_without_cycles(root, exclude_tl_dirs):
        if exclude_hidden and "/." in path:
            continue
        # path = path[2:] # strip the ./ prefix
        # if path and (path[0] == '.' or './' in path):
        #    continue
        for fname in files:
            if file_filter is None or file_filter(fname):
                p = os.path.join(path, fname)
                yield p, p[prefixlen:]


def suffix_filter(suffixes: List[str]) -> Callable[[str], bool]:
    """
    Returns a filter function that checks if a file ends with any of the given suffixes.
    """
    suffixes = [s.lower() for s in suffixes]

    def _filter(fname: str) -> bool:
        fname = fname.lower()
        return (
            suffixes is None
            or (fname[0] == "." and fname in suffixes)
            or (fname[0] != "." and any(fname.endswith(suffix) for suffix in suffixes))
        )

    return _filter


@contextmanager
def with_dir(new_dir):
    current_dir = os.getcwd()
    os.chdir(new_dir)
    yield new_dir
    os.chdir(current_dir)



import functools
import importlib
import inspect
import math
import os
import pathlib
import re
import time
import typing
from datetime import datetime
from io import StringIO
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    ForwardRef,
    Iterable,
    List,
    NewType,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from metaflow import FlowSpec, step
from metaflow.debug import debug
from metaflow.decorators import Decorator, FlowDecorator
from metaflow.extension_support import get_aliased_modules
from metaflow.metaflow_current import Current
from metaflow.metaflow_version import get_version
from metaflow.runner.deployer import DeployedFlow, Deployer, TriggeredRun
from metaflow.runner.deployer_impl import DeployerImpl

TAB = "    "
METAFLOW_CURRENT_MODULE_NAME = "metaflow.metaflow_current"
METAFLOW_DEPLOYER_MODULE_NAME = "metaflow.runner.deployer"

param_section_header = re.compile(r"Parameters\s*\n----------\s*\n", flags=re.M)
return_section_header = re.compile(r"Returns\s*\n-------\s*\n", flags=re.M)
add_to_current_header = re.compile(
    r"MF Add To Current\s*\n-----------------\s*\n", flags=re.M
)
non_indented_line = re.compile(r"^\S+.*$")
param_name_type = re.compile(r"^(?P<name>\S+)(?:\s*:\s*(?P<type>.*))?$")
type_annotations = re.compile(
    r"(?P<type>.*?)(?P<optional>, optional|\(optional\))?(?:, [Dd]efault(?: is | = |: |s to |)\s*(?P<default>.*))?$"
)

FlowSpecDerived = TypeVar("FlowSpecDerived", bound=FlowSpec)

StepFlag = NewType("StepFlag", bool)

MetaflowStepFunction = Union[
    Callable[[FlowSpecDerived, StepFlag], None],
    Callable[[FlowSpecDerived, Any, StepFlag], None],
]


# Object that has start() and end() like a Match object to make the code simpler when
# we are parsing different sections of doc
class StartEnd:
    def __init__(self, start: int, end: int):
        self._start = start
        self._end = end

    def start(self):
        return self._start

    def end(self):
        return self._end


def type_var_to_str(t: TypeVar) -> str:
    bound_name = None
    if t.__bound__ is not None:
        if isinstance(t.__bound__, typing.ForwardRef):
            bound_name = t.__bound__.__forward_arg__
        else:
            bound_name = t.__bound__.__name__
    return 'typing.TypeVar("%s", %scontravariant=%s, covariant=%s%s)' % (
        t.__name__,
        'bound="%s", ' % bound_name if t.__bound__ else "",
        t.__contravariant__,
        t.__covariant__,
        ", ".join([""] + [c.__name__ for c in t.__constraints__]),
    )


def new_type_to_str(t: typing.NewType) -> str:
    return 'typing.NewType("%s", %s)' % (t.__name__, t.__supertype__.__name__)


def descend_object(object: str, options: Iterable[str]):
    # Returns true if:
    #  - options contains a prefix of object
    #  - the component after the prefix does not start with _
    for opt in options:
        new_object = object.removeprefix(opt)
        if len(new_object) == len(object):
            # There was no prefix, so we continue
            continue
        # Using [1] to skip the inevitable "."
        if len(new_object) == 0 or new_object[1] != "_":
            return True
    return False


def parse_params_from_doc(doc: str) -> Tuple[List[inspect.Parameter], bool]:
    parameters = []
    no_arg_version = True
    for line in doc.splitlines():
        if non_indented_line.match(line):
            match = param_name_type.match(line)
            arg_name = type_name = is_optional = default = None
            default_set = False
            if match is not None:
                arg_name = match.group("name")
                type_name = match.group("type")
                if type_name is not None:
                    type_detail = type_annotations.match(type_name)
                    if type_detail is not None:
                        type_name = type_detail.group("type")
                        is_optional = type_detail.group("optional") is not None
                        default = type_detail.group("default")
                        if default:
                            default_set = True
                        try:
                            default = eval(default)
                        except:
                            pass
                        try:
                            type_name = eval(type_name)
                        except:
                            pass
                parameters.append(
                    inspect.Parameter(
                        name=arg_name,
                        kind=inspect.Parameter.KEYWORD_ONLY,
                        default=(
                            default
                            if default_set
                            else None if is_optional else inspect.Parameter.empty
                        ),
                        annotation=(Optional[type_name] if is_optional else type_name),
                    )
                )
                if not default_set:
                    # If we don't have a default set for any parameter, we can't
                    # have a no-arg version since the function would be incomplete
                    no_arg_version = False
    return parameters, no_arg_version


def split_docs(
    raw_doc: str, boundaries: List[Tuple[str, Union[StartEnd, re.Match]]]
) -> Dict[str, str]:
    docs = dict()
    boundaries.sort(key=lambda x: x[1].start())

    section_start = 0
    for idx in range(1, len(boundaries)):
        docs[boundaries[idx - 1][0]] = raw_doc[
            section_start : boundaries[idx][1].start()
        ]
        section_start = boundaries[idx][1].end()
    docs[boundaries[-1][0]] = raw_doc[section_start:]
    return docs


def parse_add_to_docs(
    raw_doc: str,
) -> Dict[str, Union[Tuple[inspect.Signature, str], str]]:
    prop = None
    return_type = None
    property_indent = None
    doc = []
    add_to_docs = dict()  # type: Dict[str, Union[str, Tuple[inspect.Signature, str]]]

    def _add():
        if prop:
            add_to_docs[prop] = (
                inspect.Signature(
                    [
                        inspect.Parameter(
                            "self", inspect.Parameter.POSITIONAL_OR_KEYWORD
                        )
                    ],
                    return_annotation=return_type,
                ),
                "\n".join(doc),
            )

    for line in raw_doc.splitlines():
        # Parse stanzas that look like the following:
        # <property-name> -> type
        # indented doc string
        if property_indent is not None and (
            line.startswith(property_indent + " ") or line.strip() == ""
        ):
            offset = len(property_indent)
            if line.lstrip().startswith("@@ "):
                line = line.replace("@@ ", "")
            doc.append(line[offset:].rstrip())
        else:
            if line.strip() == 0:
                continue
            if prop:
                # Ends a property stanza
                _add()
            # Now start a new one
            line = line.rstrip()
            property_indent = line[: len(line) - len(line.lstrip())]
            # Either this has a -> to denote a property or it is a pure name
            # to denote a reference to a function (starting with #)
            line = line.lstrip()
            if line.startswith("#"):
                # The name of the function is the last part like metaflow.deployer.run
                add_to_docs[line.split(".")[-1]] = line[1:]
                continue
            # This is a line so we split it using "->"
            prop, return_type = line.split("->")
            prop = prop.strip()
            return_type = return_type.strip()
            doc = []
    _add()
    return add_to_docs


def add_indent(indentation: str, text: str) -> str:
    return "\n".join([indentation + line for line in text.splitlines()])


class StubGenerator:
    """
    This class takes the name of a library as input and a directory as output.

    It will then generate the corresponding stub files for each defined type
    (generic variables, functions, classes, etc.) at run time.
    This means that the code for the library is not statically parsed, but it is
    executed and then the types are dynamically created and analyzed to produce the stub
    files.

    The only items analyzes are those that belong to the library (ie: anything in
    the library or below it but not any external imports)
    """

    def __init__(self, output_dir: str, include_generated_for: bool = True):
        """
        Initializes the StubGenerator.
        :param file_path: the file path
        :type file_path: str
        :param members_from_other_modules: the names of the members defined in other module to be analyzed
        :type members_from_other_modules: List[str]
        """

        # Let metaflow know we are in stubgen mode. This is sometimes useful to skip
        # some processing like loading libraries, etc. It is used in Metaflow extensions
        # so do not remove even if you do not see a use for it directly in the code.
        os.environ["METAFLOW_STUBGEN"] = "1"

        self._write_generated_for = include_generated_for
        # First element is the name it should be installed in (alias) and second is the
        # actual module name
        self._pending_modules = [
            ("metaflow", "metaflow")
        ]  # type: List[Tuple[str, str]]
        self._root_module = "metaflow."
        self._safe_modules = ["metaflow.", "metaflow_extensions."]

        self._pending_modules.extend(
            (self._get_module_name_alias(x), x) for x in get_aliased_modules()
        )

        # We exclude some modules to not create a bunch of random non-user facing
        # .pyi files.
        self._exclude_modules = set(
            [
                "metaflow.cli_args",
                "metaflow.cmd",
                "metaflow.cmd_with_io",
                "metaflow.datastore",
                "metaflow.debug",
                "metaflow.decorators",
                "metaflow.event_logger",
                "metaflow.extension_support",
                "metaflow.graph",
                "metaflow.integrations",
                "metaflow.lint",
                "metaflow.metaflow_metadata",
                "metaflow.metaflow_config_funcs",
                "metaflow.metaflow_environment",
                "metaflow.metaflow_profile",
                "metaflow.metaflow_version",
                "metaflow.mflog",
                "metaflow.monitor",
                "metaflow.package",
                "metaflow.plugins.datastores",
                "metaflow.plugins.env_escape",
                "metaflow.plugins.metadata_providers",
                "metaflow.procpoll.py",
                "metaflow.R",
                "metaflow.runtime",
                "metaflow.sidecar",
                "metaflow.task",
                "metaflow.tracing",
                "metaflow.unbounded_foreach",
                "metaflow.util",
                "metaflow._vendor",
            ]
        )

        self._done_modules = set()  # type: Set[str]
        self._output_dir = output_dir
        self._mf_version = get_version()

        # Contains the names of the methods that are injected in Deployer
        self._deployer_injected_methods = (
            {}
        )  # type: Dict[str, Dict[str, Union[Tuple[str, str], str]]]
        # Contains information to add to the Current object (injected by decorators)
        self._addl_current = (
            dict()
        )  # type: Dict[str, Dict[str, Tuple[inspect.Signature, str]]]

        self._reset()

    def _reset(self):
        # "Globals" that are used throughout processing. This is not the cleanest
        # but simplifies code quite a bit.

        # Imports that are needed at the top of the file
        self._imports = set()  # type: Set[str]

        self._sub_module_imports = set()  # type: Set[Tuple[str, str]]``
        # Typing imports (behind if TYPE_CHECKING) that are needed at the top of the file
        self._typing_imports = set()  # type: Set[str]
        # Typevars that are defined
        self._typevars = dict()  # type: Dict[str, Union[TypeVar, type]]
        # Current objects in the file being processed
        self._current_objects = {}  # type: Dict[str, Any]
        self._current_references = []  # type: List[str]
        # Current stubs in the file being processed
        self._stubs = []  # type: List[str]

        # These have a shorter "scope"
        # Current parent module of the object being processed -- used to determine
        # the "globals()"
        self._current_parent_module = None  # type: Optional[ModuleType]

    def _get_module_name_alias(self, module_name):
        if any(
            module_name.startswith(x) for x in self._safe_modules
        ) and not module_name.startswith(self._root_module):
            return self._root_module + ".".join(
                ["mf_extensions", *module_name.split(".")[1:]]
            )
        return module_name

    def _get_relative_import(
        self, new_module_name, cur_module_name, is_init_module=False
    ):
        new_components = new_module_name.split(".")
        cur_components = cur_module_name.split(".")
        init_module_count = 1 if is_init_module else 0
        common_idx = 0
        max_idx = min(len(new_components), len(cur_components))
        while (
            common_idx < max_idx
            and new_components[common_idx] == cur_components[common_idx]
        ):
            common_idx += 1
        # current: a.b and parent: a.b.e.d -> from .e.d import <name>
        # current: a.b.c.d and parent: a.b.e.f -> from ...e.f import <name>
        return "." * (len(cur_components) - common_idx + init_module_count) + ".".join(
            new_components[common_idx:]
        )

    def _get_module(self, alias, name):
        debug.stubgen_exec("Analyzing module %s (aliased at %s)..." % (name, alias))
        self._current_module = importlib.import_module(name)
        self._current_module_name = alias
        for objname, obj in self._current_module.__dict__.items():
            if objname == "_addl_stubgen_modules":
                debug.stubgen_exec(
                    "Adding modules %s from _addl_stubgen_modules" % str(obj)
                )
                self._pending_modules.extend(
                    (self._get_module_name_alias(m), m) for m in obj
                )
                continue
            if objname.startswith("_"):
                debug.stubgen_exec(
                    "Skipping object because it starts with _ %s" % objname
                )
                continue
            if inspect.ismodule(obj):
                # Only consider modules that are safe modules
                if (
                    any(obj.__name__.startswith(m) for m in self._safe_modules)
                    and not obj.__name__ in self._exclude_modules
                ):
                    debug.stubgen_exec(
                        "Adding child module %s to process" % obj.__name__
                    )

                    new_module_alias = self._get_module_name_alias(obj.__name__)
                    self._pending_modules.append((new_module_alias, obj.__name__))

                    new_parent, new_name = new_module_alias.rsplit(".", 1)
                    self._current_references.append(
                        "from %s import %s as %s"
                        % (
                            self._get_relative_import(
                                new_parent,
                                alias,
                                hasattr(self._current_module, "__path__"),
                            ),
                            new_name,
                            objname,
                        )
                    )
                else:
                    debug.stubgen_exec("Skipping child module %s" % obj.__name__)
            else:
                parent_module = inspect.getmodule(obj)
                # For objects we include:
                #  - stuff that is a functools.partial (these are all the decorators;
                #    we could be more specific but good enough for now) for root module.
                #    We also include the step decorator (it's from metaflow.decorators
                #    which is typically excluded)
                #  - Stuff that is defined in this module itself
                #  - a reference to anything in the modules we will process later
                #    (so we don't duplicate a ton of times)

                if (
                    parent_module is None
                    or (
                        name + "." == self._root_module
                        and (
                            (parent_module.__name__.startswith("functools"))
                            or obj == step
                        )
                    )
                    or parent_module.__name__ == name
                ):
                    debug.stubgen_exec("Adding object %s to process" % objname)
                    self._current_objects[objname] = obj

                elif not any(
                    [
                        parent_module.__name__.startswith(p)
                        for p in self._exclude_modules
                    ]
                ) and any(
                    [parent_module.__name__.startswith(p) for p in self._safe_modules]
                ):
                    parent_alias = self._get_module_name_alias(parent_module.__name__)

                    relative_import = self._get_relative_import(
                        parent_alias, alias, hasattr(self._current_module, "__path__")
                    )

                    debug.stubgen_exec(
                        "Adding reference %s and adding module %s as %s"
                        % (objname, parent_module.__name__, parent_alias)
                    )
                    obj_import_name = getattr(obj, "__name__", objname)
                    if obj_import_name == "<lambda>":
                        # We have one case of this
                        obj_import_name = objname
                    self._current_references.append(
                        "from %s import %s as %s"
                        % (relative_import, obj_import_name, objname)
                    )
                    self._pending_modules.append((parent_alias, parent_module.__name__))
                else:
                    debug.stubgen_exec("Skipping object %s" % objname)

    def _get_element_name_with_module(
        self, element: Union[TypeVar, type, Any], force_import=False
    ) -> str:
        # The element can be a string, for example "def f() -> 'SameClass':..."
        def _add_to_import(name):
            if name != self._current_module_name:
                self._imports.add(name)

        def _add_to_typing_check(name, is_module=False):
            if name == "None":
                return
            if is_module:
                self._typing_imports.add(name)
            else:
                splits = name.rsplit(".", 1)
                if len(splits) > 1 and not (
                    len(splits) == 2 and splits[0] == self._current_module_name
                ):
                    # We don't add things that are just one name -- probably things within
                    # the current file
                    self._typing_imports.add(splits[0])

        def _format_qualified_class_name(cls: type) -> str:
            """Helper to format a class with its qualified module name"""
            # Special case for NoneType - return None
            if cls.__name__ == "NoneType":
                return "None"

            module = inspect.getmodule(cls)
            if (
                module
                and module.__name__ != "builtins"
                and module.__name__ != "__main__"
            ):
                module_name = self._get_module_name_alias(module.__name__)
                _add_to_typing_check(module_name, is_module=True)
                return f"{module_name}.{cls.__name__}"
            else:
                return cls.__name__

        if isinstance(element, str):
            # Special case for self referential things (particularly in a class)
            if element == self._current_name:
                return '"%s"' % element
            # We first try to eval the annotation because with the annotations future
            # it is always a string
            try:
                potential_element = eval(
                    element,
                    (
                        self._current_parent_module.__dict__
                        if self._current_parent_module
                        else None
                    ),
                )
                if potential_element:
                    element = potential_element
            except:
                pass

        if isinstance(element, str):
            # If we are in our "safe" modules, make sure we alias properly
            if any(element.startswith(x) for x in self._safe_modules):
                element = self._get_module_name_alias(element)
            _add_to_typing_check(element)
            return '"%s"' % element
        # 3.10+ has NewType as a class but not before so hack around to check for NewType
        elif isinstance(element, TypeVar) or hasattr(element, "__supertype__"):
            if not element.__name__ in self._typevars:
                self._typevars[element.__name__] = element
            return element.__name__
        elif isinstance(element, type):
            module = inspect.getmodule(element)
            if (
                module is None
                or module.__name__ == "builtins"
                or module.__name__ == "__main__"
            ):
                # Special case for "NoneType" -- return None as NoneType is only 3.10+
                if element.__name__ == "NoneType":
                    return "None"
                return element.__name__

            module_name = self._get_module_name_alias(module.__name__)
            if force_import:
                _add_to_import(module_name.split(".")[0])
            _add_to_typing_check(module_name, is_module=True)
            if module_name != self._current_module_name:
                return "{0}.{1}".format(module_name, element.__name__)
            else:
                return element.__name__
        elif isinstance(element, type(Ellipsis)):
            return "..."
        elif isinstance(element, typing._GenericAlias):
            # We need to check things recursively in __args__ if it exists
            args_str = []
            for arg in getattr(element, "__args__", []):
                # Special handling for class objects in type arguments
                if isinstance(arg, type):
                    args_str.append(_format_qualified_class_name(arg))
                else:
                    args_str.append(self._get_element_name_with_module(arg))

            _add_to_import("typing")
            if element._name:
                if element._name == "Optional":
                    # We don't want to include NoneType in the string -- it breaks things
                    args_str = args_str[:1]
                elif element._name == "Callable":
                    # We need to make this a list of everything except the end one
                    # except if it is an ellipsis
                    if args_str[0] != "...":
                        call_args = "[" + ", ".join(args_str[:-1]) + "]"
                        args_str = [call_args, args_str[-1]]
                return "typing.%s[%s]" % (element._name, ", ".join(args_str))
            else:
                # Handle the case where we have a generic type without a _name
                origin = element.__origin__
                if isinstance(origin, type):
                    origin_str = _format_qualified_class_name(origin)
                else:
                    origin_str = str(origin)
                return "%s[%s]" % (origin_str, ", ".join(args_str))
        elif isinstance(element, ForwardRef):
            f_arg = self._get_module_name_alias(element.__forward_arg__)
            _add_to_typing_check(f_arg)
            return '"%s"' % f_arg
        elif inspect.getmodule(element) == inspect.getmodule(typing):
            _add_to_import("typing")
            # Special handling for NamedTuple which is a function
            if hasattr(element, "__name__") and element.__name__ == "NamedTuple":
                return "typing.NamedTuple"
            return str(element)
        else:
            if hasattr(element, "__module__"):
                elem_module = self._get_module_name_alias(element.__module__)
                if elem_module == "builtins":
                    return getattr(element, "__name__", str(element))
                _add_to_typing_check(elem_module, is_module=True)
                return "{0}.{1}".format(
                    elem_module, getattr(element, "__name__", element)
                )
            else:
                # A constant
                return str(element)

    def _exploit_annotation(self, annotation: Any, starting: str = ": ") -> str:
        annotation_string = ""
        if annotation and annotation != inspect.Parameter.empty:
            annotation_string += starting + self._get_element_name_with_module(
                annotation
            )
        return annotation_string

    def _generate_class_stub(self, name: str, clazz: type) -> str:
        debug.stubgen_exec("Generating class stub for %s" % name)
        skip_init = issubclass(clazz, (TriggeredRun, DeployedFlow))
        if issubclass(clazz, DeployerImpl):
            if clazz.TYPE is not None:
                clazz_type = clazz.TYPE.replace("-", "_")
                self._deployer_injected_methods.setdefault(clazz_type, {})[
                    "deployer"
                ] = (self._current_module_name + "." + name)

        # Handle TypedDict gracefully for Python 3.7 compatibility
        # _TypedDictMeta is not available in Python 3.7
        typed_dict_meta = getattr(typing, "_TypedDictMeta", None)
        if typed_dict_meta is not None and isinstance(clazz, typed_dict_meta):
            self._sub_module_imports.add(("typing", "TypedDict"))
            total_flag = getattr(clazz, "__total__", False)
            buff = StringIO()
            # Emit the TypedDict base and total flag
            buff.write(f"class {name}(TypedDict, total={total_flag}):\n")
            # Write out each field from __annotations__
            for field_name, field_type in clazz.__annotations__.items():
                ann = self._get_element_name_with_module(field_type)
                buff.write(f"{TAB}{field_name}: {ann}\n")
            return buff.getvalue()

        buff = StringIO()
        # Class prototype
        buff.write("class " + name.split(".")[-1] + "(")

        # Add super classes
        for c in clazz.__bases__:
            name_with_module = self._get_element_name_with_module(c, force_import=True)
            buff.write(name_with_module + ", ")

        # Add metaclass
        name_with_module = self._get_element_name_with_module(
            clazz.__class__, force_import=True
        )
        buff.write("metaclass=" + name_with_module + "):\n")

        # Add class docstring
        if clazz.__doc__:
            buff.write('%s"""\n' % TAB)
            my_doc = inspect.cleandoc(clazz.__doc__)
            init_blank = True
            for line in my_doc.split("\n"):
                if init_blank and len(line.strip()) == 0:
                    continue
                init_blank = False
                buff.write("%s%s\n" % (TAB, line.rstrip()))
            buff.write('%s"""\n' % TAB)

        # For NamedTuple, we have __annotations__ but no __init__. In that case,
        # we are going to "create" a __init__ function with the annotations
        # to show what the class takes.
        annotation_dict = None
        init_func = None
        for key, element in clazz.__dict__.items():
            func_deco = None
            if isinstance(element, staticmethod):
                func_deco = "@staticmethod"
                element = element.__func__
            elif isinstance(element, classmethod):
                func_deco = "@classmethod"
                element = element.__func__
            if key == "__init__":
                if skip_init:
                    continue
                init_func = element
            elif key == "__annotations__":
                annotation_dict = element
            if inspect.isfunction(element):
                if not element.__name__.startswith("_") or element.__name__.startswith(
                    "__"
                ):
                    if (
                        clazz == Deployer
                        and element.__name__ in self._deployer_injected_methods
                    ):
                        # This is a method that was injected. It has docs but we need
                        # to parse it to generate the proper signature
                        func_doc = inspect.cleandoc(element.__doc__)
                        docs = split_docs(
                            func_doc,
                            [
                                ("func_doc", StartEnd(0, 0)),
                                (
                                    "param_doc",
                                    param_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                                (
                                    "return_doc",
                                    return_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                            ],
                        )

                        parameters, _ = parse_params_from_doc(docs["param_doc"])
                        return_type = self._deployer_injected_methods[element.__name__][
                            "deployer"
                        ]

                        buff.write(
                            self._generate_function_stub(
                                key,
                                element,
                                sign=[
                                    inspect.Signature(
                                        parameters=[
                                            inspect.Parameter(
                                                "self",
                                                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                            )
                                        ]
                                        + parameters,
                                        return_annotation=return_type,
                                    )
                                ],
                                indentation=TAB,
                                deco=func_deco,
                            )
                        )
                    elif (
                        clazz == DeployedFlow and element.__name__ == "from_deployment"
                    ):
                        # We simply update the signature to list the return
                        # type as a union of all possible deployers
                        func_doc = inspect.cleandoc(element.__doc__)
                        docs = split_docs(
                            func_doc,
                            [
                                ("func_doc", StartEnd(0, 0)),
                                (
                                    "param_doc",
                                    param_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                                (
                                    "return_doc",
                                    return_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                            ],
                        )

                        parameters, _ = parse_params_from_doc(docs["param_doc"])

                        def _create_multi_type(*l):
                            return typing.Union[l]

                        all_types = [
                            v["from_deployment"][0]
                            for v in self._deployer_injected_methods.values()
                        ]

                        if len(all_types) > 1:
                            return_type = _create_multi_type(*all_types)
                        else:
                            return_type = all_types[0] if len(all_types) else None

                        buff.write(
                            self._generate_function_stub(
                                key,
                                element,
                                sign=[
                                    inspect.Signature(
                                        parameters=[
                                            inspect.Parameter(
                                                "cls",
                                                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                            )
                                        ]
                                        + parameters,
                                        return_annotation=return_type,
                                    )
                                ],
                                indentation=TAB,
                                doc=docs["func_doc"]
                                + "\n\nParameters\n----------\n"
                                + docs["param_doc"]
                                + "\n\nReturns\n-------\n"
                                + "%s\nA `DeployedFlow` object" % str(return_type),
                                deco=func_deco,
                            )
                        )
                    elif (
                        clazz == DeployedFlow
                        and element.__name__.startswith("from_")
                        and element.__name__[5:] in self._deployer_injected_methods
                    ):
                        # Get the doc from the from_deployment method stored in
                        # self._deployer_injected_methods
                        func_doc = inspect.cleandoc(
                            self._deployer_injected_methods[element.__name__[5:]][
                                "from_deployment"
                            ][1]
                            or ""
                        )
                        docs = split_docs(
                            func_doc,
                            [
                                ("func_doc", StartEnd(0, 0)),
                                (
                                    "param_doc",
                                    param_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                                (
                                    "return_doc",
                                    return_section_header.search(func_doc)
                                    or StartEnd(len(func_doc), len(func_doc)),
                                ),
                            ],
                        )

                        parameters, _ = parse_params_from_doc(docs["param_doc"])
                        return_type = self._deployer_injected_methods[
                            element.__name__[5:]
                        ]["from_deployment"][0]

                        buff.write(
                            self._generate_function_stub(
                                key,
                                element,
                                sign=[
                                    inspect.Signature(
                                        parameters=[
                                            inspect.Parameter(
                                                "cls",
                                                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                            )
                                        ]
                                        + parameters,
                                        return_annotation=return_type,
                                    )
                                ],
                                indentation=TAB,
                                doc=docs["func_doc"]
                                + "\n\nParameters\n----------\n"
                                + docs["param_doc"]
                                + "\n\nReturns\n-------\n"
                                + docs["return_doc"],
                                deco=func_deco,
                            )
                        )
                    else:
                        if (
                            issubclass(clazz, DeployedFlow)
                            and clazz.TYPE is not None
                            and key == "from_deployment"
                        ):
                            clazz_type = clazz.TYPE.replace("-", "_")
                            # Record docstring for this function
                            self._deployer_injected_methods.setdefault(clazz_type, {})[
                                "from_deployment"
                            ] = (
                                self._current_module_name + "." + name,
                                element.__doc__,
                            )
                        buff.write(
                            self._generate_function_stub(
                                key,
                                element,
                                indentation=TAB,
                                deco=func_deco,
                            )
                        )

            elif isinstance(element, property):
                if element.fget:
                    buff.write(
                        self._generate_function_stub(
                            key, element.fget, indentation=TAB, deco="@property"
                        )
                    )
                if element.fset:
                    buff.write(
                        self._generate_function_stub(
                            key, element.fset, indentation=TAB, deco="@%s.setter" % key
                        )
                    )

        # Special handling of classes that have injected methods
        if clazz == Current:
            # Multiple decorators can add the same object (trigger and trigger_on_finish)
            # as examples so we sort it out.
            resulting_dict = (
                dict()
            )  # type Dict[str, List[inspect.Signature, str, List[str]]]
            for deco_name, addl_current in self._addl_current.items():
                for name, (sign, doc) in addl_current.items():
                    r = resulting_dict.setdefault(name, [sign, doc, []])
                    r[2].append("@%s" % deco_name)
            for name, (sign, doc, decos) in resulting_dict.items():
                buff.write(
                    self._generate_function_stub(
                        name,
                        sign=[sign],
                        indentation=TAB,
                        doc="(only in the presence of the %s decorator%s)\n\n"
                        % (", or ".join(decos), "" if len(decos) == 1 else "s")
                        + doc,
                        deco="@property",
                    )
                )

        if not skip_init and init_func is None and annotation_dict:
            buff.write(
                self._generate_function_stub(
                    "__init__",
                    func=None,
                    sign=[
                        inspect.Signature(
                            parameters=[
                                inspect.Parameter(
                                    name="self",
                                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                )
                            ]
                            + [
                                inspect.Parameter(
                                    name=name,
                                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                    annotation=annotation,
                                )
                                for name, annotation in annotation_dict.items()
                            ]
                        )
                    ],
                    indentation=TAB,
                )
            )
        buff.write("%s...\n" % TAB)

        return buff.getvalue()

    def _extract_signature_from_decorator(
        self, name: str, raw_doc: Optional[str], is_flow_decorator: bool = False
    ) -> Optional[List[Tuple[inspect.Signature, str]]]:
        # TODO: This only handles the `Parameters` section for now; we are
        # using it only to parse the documentation for step/flow decorators so
        # this is enough for now but it could be extended more.
        # Inspired from:
        # https://github.com/rr-/docstring_parser/blob/master/docstring_parser/numpydoc.py
        if raw_doc is None:
            return None

        if not "FlowSpecDerived" in self._typevars:
            self._typevars["FlowSpecDerived"] = FlowSpecDerived
            self._typevars["StepFlag"] = StepFlag

        raw_doc = inspect.cleandoc(raw_doc)
        section_boundaries = [
            ("func_doc", StartEnd(0, 0)),
            (
                "param_doc",
                param_section_header.search(raw_doc)
                or StartEnd(len(raw_doc), len(raw_doc)),
            ),
            (
                "add_to_current_doc",
                add_to_current_header.search(raw_doc)
                or StartEnd(len(raw_doc), len(raw_doc)),
            ),
        ]

        docs = split_docs(raw_doc, section_boundaries)
        parameters, no_arg_version = parse_params_from_doc(docs["param_doc"])

        if docs["add_to_current_doc"]:
            self._addl_current[name] = parse_add_to_docs(docs["add_to_current_doc"])

        result = []
        if no_arg_version:
            if is_flow_decorator:
                if docs["param_doc"]:
                    result.append(
                        (
                            inspect.Signature(
                                parameters=parameters,
                                return_annotation=Callable[
                                    [typing.Type[FlowSpecDerived]],
                                    typing.Type[FlowSpecDerived],
                                ],
                            ),
                            "",
                        )
                    )
                result.append(
                    (
                        inspect.Signature(
                            parameters=[
                                inspect.Parameter(
                                    name="f",
                                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                    annotation=typing.Type[FlowSpecDerived],
                                )
                            ],
                            return_annotation=typing.Type[FlowSpecDerived],
                        ),
                        "",
                    ),
                )
            else:
                if docs["param_doc"]:
                    result.append(
                        (
                            inspect.Signature(
                                parameters=parameters,
                                return_annotation=typing.Callable[
                                    [MetaflowStepFunction], MetaflowStepFunction
                                ],
                            ),
                            "",
                        )
                    )
                result.extend(
                    [
                        (
                            inspect.Signature(
                                parameters=[
                                    inspect.Parameter(
                                        name="f",
                                        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                        annotation=Callable[
                                            [FlowSpecDerived, StepFlag], None
                                        ],
                                    )
                                ],
                                return_annotation=Callable[
                                    [FlowSpecDerived, StepFlag], None
                                ],
                            ),
                            "",
                        ),
                        (
                            inspect.Signature(
                                parameters=[
                                    inspect.Parameter(
                                        name="f",
                                        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                        annotation=Callable[
                                            [FlowSpecDerived, Any, StepFlag], None
                                        ],
                                    )
                                ],
                                return_annotation=Callable[
                                    [FlowSpecDerived, Any, StepFlag], None
                                ],
                            ),
                            "",
                        ),
                    ]
                )

        if is_flow_decorator:
            result = result + [
                (
                    inspect.Signature(
                        parameters=(
                            [
                                inspect.Parameter(
                                    name="f",
                                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                    annotation=Optional[typing.Type[FlowSpecDerived]],
                                    default=(
                                        None
                                        if no_arg_version
                                        else inspect.Parameter.empty
                                    ),
                                )
                            ]
                            + parameters
                            if no_arg_version
                            else [] + parameters
                        ),
                        return_annotation=(
                            inspect.Signature.empty
                            if no_arg_version
                            else Callable[
                                [typing.Type[FlowSpecDerived]],
                                typing.Type[FlowSpecDerived],
                            ]
                        ),
                    ),
                    "",
                ),
            ]
        else:
            result = result + [
                (
                    inspect.Signature(
                        parameters=(
                            [
                                inspect.Parameter(
                                    name="f",
                                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                    annotation=Optional[MetaflowStepFunction],
                                    default=(
                                        None
                                        if no_arg_version
                                        else inspect.Parameter.empty
                                    ),
                                )
                            ]
                            + parameters
                            if no_arg_version
                            else [] + parameters
                        ),
                        return_annotation=(
                            inspect.Signature.empty
                            if no_arg_version
                            else typing.Callable[
                                [MetaflowStepFunction], MetaflowStepFunction
                            ]
                        ),
                    ),
                    "",
                ),
            ]
        if len(result) == 2:
            # If we only have one overload -- we don't need it at all. Happens for
            # flow-level decorators that don't take any arguments
            result = result[1:]
        # Add doc to first and last overloads. Jedi uses the last one and pycharm
        # the first one. Go figure.
        result_docstring = docs["func_doc"]
        if docs["param_doc"]:
            result_docstring += "\nParameters\n----------\n" + docs["param_doc"]
        result[0] = (
            result[0][0],
            result_docstring,
        )
        result[-1] = (
            result[-1][0],
            result_docstring,
        )
        return result

    def _generate_function_stub(
        self,
        name: str,
        func: Optional[Union[Callable, classmethod]] = None,
        sign: Optional[List[inspect.Signature]] = None,
        indentation: Optional[str] = None,
        doc: Optional[str] = None,
        deco: Optional[str] = None,
    ) -> str:
        debug.stubgen_exec("Generating function stub for %s" % name)

        def exploit_default(default_value: Any) -> Optional[str]:
            if default_value == inspect.Parameter.empty:
                return None
            if type(default_value).__module__ == "builtins":
                if isinstance(default_value, list):
                    return (
                        "["
                        + ", ".join(
                            [cast(str, exploit_default(v)) for v in default_value]
                        )
                        + "]"
                    )
                elif isinstance(default_value, tuple):
                    return (
                        "("
                        + ", ".join(
                            [cast(str, exploit_default(v)) for v in default_value]
                        )
                        + ")"
                    )
                elif isinstance(default_value, dict):
                    return (
                        "{"
                        + ", ".join(
                            [
                                cast(str, exploit_default(k))
                                + ": "
                                + cast(str, exploit_default(v))
                                for k, v in default_value.items()
                            ]
                        )
                        + "}"
                    )
                elif isinstance(default_value, str):
                    return repr(default_value)  # Use repr() for proper escaping
                elif isinstance(default_value, (int, float, bool)):
                    return str(default_value)
                elif default_value is None:
                    return "None"
                else:
                    return "..."  # For other built-in types not explicitly handled
            elif inspect.isclass(default_value) or inspect.isfunction(default_value):
                if default_value.__module__ == "builtins":
                    return default_value.__name__
                else:
                    self._typing_imports.add(default_value.__module__)
                    return ".".join([default_value.__module__, default_value.__name__])
            else:
                return "..."  # For complex objects like class instances

        buff = StringIO()
        if sign is None and func is None:
            raise RuntimeError(
                "Cannot generate stub for function %s with either a function or signature"
                % name
            )
        try:
            sign = sign or [inspect.signature(cast(Callable, func))]
        except ValueError:
            # In 3.7, NamedTuples have properties that then give an operator.itemgetter
            # which doesn't have a signature. We ignore for now. It doesn't have much
            # value
            return ""
        doc = doc or func.__doc__
        if doc == "STUBGEN_IGNORE":
            # Ignore methods that have STUBGEN_IGNORE. Used to ignore certain
            # methods for the Deployer
            return ""
        indentation = indentation or ""

        # Deal with overload annotations -- the last one will be non overloaded and
        # will be the one that shows up as the type hint (for Jedi and PyCharm which
        # don't handle overloads as well)
        do_overload = False
        if sign and len(sign) > 1:
            do_overload = True
        for count, my_sign in enumerate(sign):
            if count > 0:
                buff.write("\n")

            if do_overload and count < len(sign) - 1:
                # According to mypy, we should have this on all variants but
                # some IDEs seem to prefer if there is one non-overloaded
                # This also changes our checks so if changing, modify tests
                buff.write(indentation + "@typing.overload\n")
            if deco:
                buff.write(indentation + deco + "\n")
            buff.write(indentation + "def " + name + "(")
            kw_only_param = False
            has_var_args = False
            for i, (par_name, parameter) in enumerate(my_sign.parameters.items()):
                annotation = self._exploit_annotation(parameter.annotation)
                default = exploit_default(parameter.default)

                if (
                    kw_only_param
                    and not has_var_args
                    and parameter.kind != inspect.Parameter.KEYWORD_ONLY
                ):
                    raise RuntimeError(
                        "In function '%s': cannot have a positional parameter after a "
                        "keyword only parameter" % name
                    )

                if (
                    parameter.kind == inspect.Parameter.KEYWORD_ONLY
                    and not kw_only_param
                    and not has_var_args
                ):
                    kw_only_param = True
                    buff.write("*, ")
                if parameter.kind == inspect.Parameter.VAR_KEYWORD:
                    par_name = "**%s" % par_name
                elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
                    has_var_args = True
                    par_name = "*%s" % par_name

                if default:
                    buff.write(par_name + annotation + " = " + default)
                else:
                    buff.write(par_name + annotation)

                if i < len(my_sign.parameters) - 1:
                    buff.write(", ")
            ret_annotation = self._exploit_annotation(
                my_sign.return_annotation, starting=" -> "
            )
            buff.write(")" + ret_annotation + ":\n")

            if (count == 0 or count == len(sign) - 1) and doc is not None:
                buff.write('%s%s"""\n' % (indentation, TAB))
                my_doc = inspect.cleandoc(doc)
                init_blank = True
                for line in my_doc.split("\n"):
                    if init_blank and len(line.strip()) == 0:
                        continue
                    init_blank = False
                    buff.write("%s%s%s\n" % (indentation, TAB, line.rstrip()))
                buff.write('%s%s"""\n' % (indentation, TAB))
            buff.write("%s%s...\n" % (indentation, TAB))
        return buff.getvalue()

    def _generate_generic_stub(self, element_name: str, element: Any) -> str:
        return "{0}: {1}\n".format(
            element_name, self._get_element_name_with_module(type(element))
        )

    def _generate_stubs(self):
        for name, attr in self._current_objects.items():
            self._current_parent_module = inspect.getmodule(attr)
            self._current_name = name
            if inspect.isclass(attr):
                self._stubs.append(self._generate_class_stub(name, attr))
            elif inspect.isfunction(attr):
                # Special handling of the `step` function where we want to add an
                # overload. This is just a single case so we don't make it general.
                # Unfortunately, when iterating, it doesn't see the @overload
                if (
                    name == "step"
                    and self._current_module_name == self._root_module[:-1]
                ):
                    self._stubs.append(
                        self._generate_function_stub(
                            name,
                            func=attr,
                            sign=[
                                inspect.Signature(
                                    parameters=[
                                        inspect.Parameter(
                                            name="f",
                                            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                            annotation=Callable[
                                                [FlowSpecDerived], None
                                            ],
                                        )
                                    ],
                                    return_annotation=Callable[
                                        [FlowSpecDerived, StepFlag], None
                                    ],
                                ),
                                inspect.Signature(
                                    parameters=[
                                        inspect.Parameter(
                                            name="f",
                                            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                            annotation=Callable[
                                                [FlowSpecDerived, Any], None
                                            ],
                                        )
                                    ],
                                    return_annotation=Callable[
                                        [FlowSpecDerived, Any, StepFlag], None
                                    ],
                                ),
                                inspect.signature(attr),
                            ],
                        )
                    )
                else:
                    self._stubs.append(self._generate_function_stub(name, attr))
            elif isinstance(attr, functools.partial):
                if issubclass(attr.args[0], Decorator):
                    # Special case where we are going to extract the parameters from
                    # the docstring to make the decorator look nicer
                    res = self._extract_signature_from_decorator(
                        name,
                        attr.args[0].__doc__,
                        is_flow_decorator=issubclass(attr.args[0], FlowDecorator),
                    )
                    if res:
                        self._stubs.append(
                            self._generate_function_stub(
                                name,
                                func=attr.func,
                                sign=[r[0] for r in res],
                                doc=res[-1][1],
                            )
                        )
                    else:
                        # print(
                        #    "WARNING: Could not extract decorator signature for %s"
                        #    % name
                        # )
                        pass
                else:
                    self._stubs.append(
                        self._generate_function_stub(
                            name, attr.func, doc=attr.args[0].__doc__
                        )
                    )
            elif not inspect.ismodule(attr):
                self._stubs.append(self._generate_generic_stub(name, attr))

    def _write_header(self, f, width):
        title_line = "Auto-generated Metaflow stub file"
        title_white_space = (width - len(title_line)) / 2
        title_line = "#%s%s%s#\n" % (
            " " * math.floor(title_white_space),
            title_line,
            " " * math.ceil(title_white_space),
        )
        f.write(
            "#" * (width + 2)
            + "\n"
            + title_line
            + "# MF version: %s%s#\n"
            % (self._mf_version, " " * (width - 13 - len(self._mf_version)))
            + "# Generated on %s%s#\n"
            % (
                datetime.fromtimestamp(time.time()).isoformat(),
                " " * (width - 14 - 26),
            )
            + "#" * (width + 2)
            + "\n\n"
        )

    def write_out(self):
        out_dir = self._output_dir
        os.makedirs(out_dir, exist_ok=True)
        # Write out py.typed (pylance seems to require it even though it is not
        # required in PEP 561) as well as a file we will use to check the "version"
        # of the stubs -- this helps to inform the user if the stubs were generated
        # for another version of Metaflow.
        pathlib.Path(os.path.join(out_dir, "py.typed")).touch()
        if self._write_generated_for:
            pathlib.Path(os.path.join(out_dir, "generated_for.txt")).write_text(
                "%s %s"
                % (self._mf_version, datetime.fromtimestamp(time.time()).isoformat())
            )
        post_process_modules = []
        is_post_processing = False
        while len(self._pending_modules) != 0 or len(post_process_modules) != 0:
            if is_post_processing or len(self._pending_modules) == 0:
                is_post_processing = True
                module_alias, module_name = post_process_modules.pop(0)
            else:
                module_alias, module_name = self._pending_modules.pop(0)
            # Skip vendored stuff
            if module_alias.startswith("metaflow._vendor") or module_name.startswith(
                "metaflow._vendor"
            ):
                continue
            # We delay current module and deployer module to the end since they
            # depend on info we gather elsewhere
            if (
                module_alias
                in (
                    METAFLOW_CURRENT_MODULE_NAME,
                    METAFLOW_DEPLOYER_MODULE_NAME,
                )
                and len(self._pending_modules) != 0
            ):
                post_process_modules.append((module_alias, module_name))
                continue
            if module_alias in self._done_modules:
                continue
            self._done_modules.add(module_alias)
            # If not, we process the module
            self._reset()
            self._get_module(module_alias, module_name)
            if module_name == "metaflow" and not is_post_processing:
                # We will want to regenerate this at the end to take into account
                # any changes to the Deployer
                post_process_modules.append((module_name, module_name))
                self._done_modules.remove(module_name)
                continue
            self._generate_stubs()

            if hasattr(self._current_module, "__path__"):
                # This is a package (so a directory) and we are dealing with
                # a __init__.pyi type of case
                dir_path = os.path.join(self._output_dir, *module_alias.split(".")[1:])
            else:
                # This is NOT a package so the original source file is not a __init__.py
                dir_path = os.path.join(
                    self._output_dir, *module_alias.split(".")[1:-1]
                )
            out_file = os.path.join(
                dir_path, os.path.basename(self._current_module.__file__) + "i"
            )

            width = 100

            os.makedirs(os.path.dirname(out_file), exist_ok=True)
            # We want to make sure we always have a __init__.pyi in the directories
            # we are creating
            parts = dir_path.split(os.sep)[len(self._output_dir.split(os.sep)) :]
            for i in range(1, len(parts) + 1):
                init_file_path = os.path.join(
                    self._output_dir, *parts[:i], "__init__.pyi"
                )
                if not os.path.exists(init_file_path):
                    with open(init_file_path, mode="w", encoding="utf-8") as f:
                        self._write_header(f, width)

            with open(out_file, mode="w", encoding="utf-8") as f:
                self._write_header(f, width)

                f.write("from __future__ import annotations\n\n")
                imported_typing = False
                for module in self._imports:
                    f.write("import " + module + "\n")
                    if module == "typing":
                        imported_typing = True
                for module, sub_module in self._sub_module_imports:
                    f.write(f"from {module} import {sub_module}\n")
                if self._typing_imports:
                    if not imported_typing:
                        f.write("import typing\n")
                        imported_typing = True
                    f.write("if typing.TYPE_CHECKING:\n")
                    for module in self._typing_imports:
                        f.write(TAB + "import " + module + "\n")
                if self._typevars:
                    if not imported_typing:
                        f.write("import typing\n")
                        imported_typing = True
                    for type_name, type_var in self._typevars.items():
                        if isinstance(type_var, TypeVar):
                            f.write(
                                "%s = %s\n" % (type_name, type_var_to_str(type_var))
                            )
                        else:
                            f.write(
                                "%s = %s\n" % (type_name, new_type_to_str(type_var))
                            )
                f.write("\n")
                for import_line in self._current_references:
                    f.write(import_line + "\n")
                f.write("\n")
                for stub in self._stubs:
                    f.write(stub + "\n")
            if is_post_processing:
                # Don't consider any pending modules if we are post processing
                self._pending_modules.clear()


if __name__ == "__main__":
    gen = StubGenerator("./stubs")
    gen.write_out()



from __future__ import print_function
from datetime import datetime
import random
import time
import sys
import os

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    DATATOOLS_CLIENT_PARAMS,
    DATATOOLS_SESSION_VARS,
    S3_RETRY_COUNT,
    RETRY_WARNING_THRESHOLD,
)


TEST_S3_RETRY = "TEST_S3_RETRY" in os.environ

TRANSIENT_RETRY_LINE_CONTENT = "<none>"
TRANSIENT_RETRY_START_LINE = "### RETRY INPUTS ###"


def get_s3_client(s3_role_arn=None, s3_session_vars=None, s3_client_params=None):
    from metaflow.plugins.aws.aws_client import get_aws_client

    return get_aws_client(
        "s3",
        with_error=True,
        role_arn=s3_role_arn,
        session_vars=s3_session_vars if s3_session_vars else DATATOOLS_SESSION_VARS,
        client_params=s3_client_params if s3_client_params else DATATOOLS_CLIENT_PARAMS,
    )


# decorator to retry functions that access S3
def aws_retry(f):
    def retry_wrapper(self, *args, **kwargs):
        last_exc = None
        for i in range(S3_RETRY_COUNT + 1):
            try:
                ret = f(self, *args, **kwargs)
                if TEST_S3_RETRY and i == 0:
                    raise Exception(
                        "TEST_S3_RETRY env var set. "
                        "Pretending that an S3 op failed. "
                        "This is not a real failure."
                    )
                else:
                    return ret
            except MetaflowException as ex:
                # MetaflowExceptions are not related to AWS, don't retry
                raise
            except Exception as ex:
                try:
                    function_name = f.func_name
                except AttributeError:
                    function_name = f.__name__
                if TEST_S3_RETRY and i == 0:
                    # This is applicable when this code is being tested
                    sys.stderr.write(
                        "[WARNING] S3 datastore operation %s failed (%s). "
                        "Retrying %d more times..\n"
                        % (function_name, ex, S3_RETRY_COUNT - i)
                    )
                if i + 1 > RETRY_WARNING_THRESHOLD:
                    # In a real failure, print this warning message only after a certain
                    # amount of retries
                    sys.stderr.write(
                        "[WARNING] S3 datastore operation %s failed (%s). "
                        "Retrying %d more times..\n"
                        % (function_name, ex, S3_RETRY_COUNT - i)
                    )
                self.reset_client(hard_reset=True)
                last_exc = ex
                # exponential backoff for real failures
                if not (TEST_S3_RETRY and i == 0):
                    time.sleep(2**i + random.randint(0, 5))
        raise last_exc

    return retry_wrapper


# Read an AWS source in a chunked manner.
# We read in chunks (at most 2GB -- here this is passed via max_chunk_size)
# because of https://bugs.python.org/issue42853 (Py3 bug); this also helps
# keep memory consumption lower
# NOTE: For some weird reason, if you pass a large value to
# read it delays the call, so we always pass it either what
# remains or 2GB, whichever is smallest.
def read_in_chunks(dst, src, src_sz, max_chunk_size):
    remaining = src_sz
    while remaining > 0:
        buf = src.read(min(remaining, max_chunk_size))
        # Py2 doesn't return the number of bytes written so calculate size
        # separately
        dst.write(buf)
        remaining -= len(buf)


def get_timestamp(dt):
    """
    Python2 compatible way to compute the timestamp (seconds since 1/1/1970)
    """
    return (dt.replace(tzinfo=None) - datetime(1970, 1, 1)).total_seconds()



import importlib
import json
import os
from collections.abc import Mapping
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from metaflow._vendor import click
from metaflow.debug import debug

from .config_parameters import ConfigValue
from ..exception import MetaflowException, MetaflowInternalError
from ..packaging_sys import MetaflowCodeContent
from ..parameters import DeployTimeField, ParameterContext, current_flow
from ..util import get_username


_CONVERT_PREFIX = "@!c!@:"
_DEFAULT_PREFIX = "@!d!@:"
_NO_FILE = "@!n!@:"

_CONVERTED_DEFAULT = _CONVERT_PREFIX + _DEFAULT_PREFIX
_CONVERTED_NO_FILE = _CONVERT_PREFIX + _NO_FILE
_CONVERTED_DEFAULT_NO_FILE = _CONVERTED_DEFAULT + _NO_FILE


def _load_config_values(info_file: Optional[str] = None) -> Optional[Dict[Any, Any]]:
    if info_file is None:
        config_content = MetaflowCodeContent.get_config()
    else:
        try:
            with open(info_file, encoding="utf-8") as f:
                config_content = json.load(f)
        except IOError:
            return None
    if config_content:
        return config_content.get("user_configs", {})
    return None


class ConvertPath(click.Path):
    name = "ConvertPath"

    def convert(self, value, param, ctx):
        if isinstance(value, str) and value.startswith(_CONVERT_PREFIX):
            return value
        is_default = False
        if value and value.startswith(_DEFAULT_PREFIX):
            is_default = True
            value = value[len(_DEFAULT_PREFIX) :]
        value = super().convert(value, param, ctx)
        return self.convert_value(value, is_default)

    @staticmethod
    def mark_as_default(value):
        if value is None:
            return None
        return _DEFAULT_PREFIX + str(value)

    @staticmethod
    def convert_value(value, is_default):
        default_str = _DEFAULT_PREFIX if is_default else ""
        if value is None:
            return None
        try:
            with open(value, "r", encoding="utf-8") as f:
                content = f.read()
        except OSError:
            return _CONVERT_PREFIX + default_str + _NO_FILE + value
        return _CONVERT_PREFIX + default_str + content


class ConvertDictOrStr(click.ParamType):
    name = "ConvertDictOrStr"

    def convert(self, value, param, ctx):
        is_default = False
        if isinstance(value, str):
            if value.startswith(_CONVERT_PREFIX):
                return value
            if value.startswith(_DEFAULT_PREFIX):
                is_default = True
                value = value[len(_DEFAULT_PREFIX) :]

        return self.convert_value(value, is_default)

    @staticmethod
    def convert_value(value, is_default):
        default_str = _DEFAULT_PREFIX if is_default else ""
        if value is None:
            return None

        if isinstance(value, dict):
            return _CONVERT_PREFIX + default_str + json.dumps(value)

        if value.startswith(_CONVERT_PREFIX):
            return value

        return _CONVERT_PREFIX + default_str + value

    @staticmethod
    def mark_as_default(value):
        if value is None:
            return None
        if isinstance(value, dict):
            return _DEFAULT_PREFIX + json.dumps(value)
        return _DEFAULT_PREFIX + str(value)


class MultipleTuple(click.Tuple):
    # Small wrapper around a click.Tuple to allow the environment variable for
    # configurations to be a JSON string. Otherwise the default behavior is splitting
    # by whitespace which is totally not what we want
    # You can now pass multiple configuration options through an environment variable
    # using something like:
    # METAFLOW_FLOW_CONFIG_VALUE='{"config1": {"key0": "value0"}, "config2": {"key1": "value1"}}'
    # or METAFLOW_FLOW_CONFIG='{"config1": "file1", "config2": "file2"}'

    def split_envvar_value(self, rv):
        loaded = json.loads(rv)
        return list(
            item if isinstance(item, str) else json.dumps(item)
            for pair in loaded.items()
            for item in pair
        )


class ConfigInput:
    # ConfigInput is an internal class responsible for processing all the --config and
    # --config-value options.
    # It gathers information from the --local-config-file (to figure out
    # where options are stored) and is also responsible for processing any `--config` or
    # `--config-value` options. Note that the process_configs function will be called
    # *twice* (once for the configs and another for the config-values). This makes
    # this function a little bit more tricky. We need to wait for both calls before
    # being able to process anything.

    # It will then store this information in the flow spec for use later in processing.
    # It is stored in the flow spec to avoid being global to support the Runner.

    loaded_configs = None  # type: Optional[Dict[str, Dict[Any, Any]]]
    config_file = None  # type: Optional[str]

    def __init__(
        self,
        req_configs: List[str],
        defaults: Dict[str, Tuple[Union[str, Dict[Any, Any]], bool]],
        parsers: Dict[str, Tuple[Union[str, Callable[[str], Dict[Any, Any]]], bool]],
    ):
        self._req_configs = set(req_configs)
        self._defaults = defaults
        self._parsers = parsers
        self._path_values = None
        self._value_values = None

    @staticmethod
    def make_key_name(name: str) -> str:
        # Special mark to indicate that the configuration value is not content or a file
        # name but a value that should be read in the config file (effectively where
        # the value has already been materialized).
        return "kv." + name

    @classmethod
    def set_config_file(cls, config_file: str):
        cls.config_file = config_file

    @classmethod
    def get_config(cls, config_name: str) -> Optional[Dict[Any, Any]]:
        if cls.loaded_configs is None:
            all_configs = _load_config_values(cls.config_file)
            if all_configs is None:
                raise MetaflowException(
                    "Could not load expected configuration values "
                    "from the CONFIG_PARAMETERS file. This is a Metaflow bug. "
                    "Please contact support."
                )
            cls.loaded_configs = all_configs
        return cls.loaded_configs[config_name]

    def process_configs(
        self,
        flow_name: str,
        param_name: str,
        param_value: Dict[str, Optional[str]],
        quiet: bool,
        datastore: str,
        click_obj: Optional[Any] = None,
    ):
        from ..cli import echo_always, echo_dev_null  # Prevent circular import
        from ..flowspec import FlowStateItems  # Prevent circular import

        flow_cls = getattr(current_flow, "flow_cls", None)
        if flow_cls is None:
            # This is an error
            raise MetaflowInternalError(
                "Config values should be processed for a FlowSpec"
            )

        # This function is called by click when processing all the --config and
        # --config-value options.
        # The value passed in is a list of tuples (name, value).
        # Click will provide:
        #   - all the defaults if nothing is provided on the command line
        #   - provide *just* the passed in value if anything is provided on the command
        #     line.
        #
        # We need to get all config and config-value options and click will call this
        # function twice. We will first get all the values on the command line and
        # *then* merge with the defaults to form a full set of values.
        # We therefore get a full set of values where:
        #  - the name will correspond to the configuration name
        #  - the value will be:
        #       - the default (including None if there is no default). If the default is
        #         not None, it will start with _CONVERTED_DEFAULT since Click will make
        #         the value go through ConvertPath or ConvertDictOrStr
        #       - the actual value passed through prefixed with _CONVERT_PREFIX

        debug.userconf_exec(
            "Processing configs for %s -- incoming values: %s"
            % (param_name, str(param_value))
        )

        do_return = self._value_values is None and self._path_values is None
        # We only keep around non default values. We could simplify by checking just one
        # value and if it is default it means all are but this doesn't seem much more effort
        # and is clearer
        if param_name == "config_value":
            self._value_values = {
                k: v
                for k, v in param_value.items()
                if v is not None and not v.startswith(_CONVERTED_DEFAULT)
            }
        else:
            self._path_values = {
                k: v
                for k, v in param_value.items()
                if v is not None and not v.startswith(_CONVERTED_DEFAULT)
            }
        if do_return:
            # One of values["value"] or values["path"] is None -- we are in the first
            # go around
            debug.userconf_exec("Incomplete config options; waiting for more")
            return None

        # The second go around, we process all the values and merge them.

        # If we are processing options that start with kv., we know we are in a subprocess
        # and ignore other stuff. In particular, environment variables used to pass
        # down configurations (like METAFLOW_FLOW_CONFIG) could still be present and
        # would cause an issue -- we can ignore those as the kv. values should trump
        # everything else.
        # NOTE: These are all *non default* keys
        all_keys = set(self._value_values).union(self._path_values)

        if all_keys and click_obj:
            click_obj.has_cl_config_options = True
        # Make sure we have at least some non default keys (we need some if we have
        # all kv)
        has_all_kv = all_keys and all(
            self._value_values.get(k, "").startswith(_CONVERT_PREFIX + "kv.")
            for k in all_keys
        )

        to_return = {}

        if not has_all_kv:
            # Check that the user didn't provide *both* a path and a value. Again, these
            # are only user-provided (not defaults)
            common_keys = set(self._value_values or []).intersection(
                [k for k, v in self._path_values.items()] or []
            )
            if common_keys:
                exc = click.UsageError(
                    "Cannot provide both a value and a file for the same configuration. "
                    "Found such values for '%s'" % "', '".join(common_keys)
                )
                if click_obj:
                    click_obj.delayed_config_exception = exc
                    return None
                raise exc

            all_values = dict(self._path_values)
            all_values.update(self._value_values)

            debug.userconf_exec("All config values: %s" % str(all_values))

            merged_configs = {}
            # Now look at everything (including defaults)
            for name, (val, is_path) in self._defaults.items():
                n = name
                if n in all_values:
                    # We have the value provided by the user -- use that.
                    merged_configs[n] = all_values[n]
                else:
                    # No value provided by the user -- use the default
                    if isinstance(val, DeployTimeField):
                        # This supports a default value that is a deploy-time field (similar
                        # to Parameter).)
                        # We will form our own context and pass it down -- note that you cannot
                        # use configs in the default value of configs as this introduces a bit
                        # of circularity. Note also that quiet and datastore are *eager*
                        # options so are available here.
                        param_ctx = ParameterContext(
                            flow_name=flow_name,
                            user_name=get_username(),
                            parameter_name=n,
                            logger=(echo_dev_null if quiet else echo_always),
                            ds_type=datastore,
                            configs=None,
                        )
                        val = val.fun(param_ctx)
                    if is_path:
                        # This is a file path
                        merged_configs[n] = ConvertPath.convert_value(val, True)
                    else:
                        # This is a value
                        merged_configs[n] = ConvertDictOrStr.convert_value(val, True)
        else:
            debug.userconf_exec("Fast path due to pre-processed values")
            merged_configs = self._value_values

        if click_obj:
            click_obj.has_config_options = True

        debug.userconf_exec("Configs merged with defaults: %s" % str(merged_configs))

        missing_configs = set()
        no_file = []
        no_default_file = []
        msgs = []
        for name, val in merged_configs.items():
            if val is None:
                missing_configs.add(name)
                to_return[name] = None
                flow_cls._flow_state.self_data[FlowStateItems.CONFIGS][name] = (
                    None,
                    self._parsers[name][1],
                )
                continue
            if val.startswith(_CONVERTED_NO_FILE):
                no_file.append(name)
                continue
            if val.startswith(_CONVERTED_DEFAULT_NO_FILE):
                no_default_file.append(name)
                continue

            parser, is_plain = self._parsers[name]
            val = val[len(_CONVERT_PREFIX) :]  # Remove the _CONVERT_PREFIX
            if val.startswith(_DEFAULT_PREFIX):  # Remove the _DEFAULT_PREFIX if needed
                val = val[len(_DEFAULT_PREFIX) :]
            if val.startswith("kv."):
                # This means to load it from a file
                try:
                    read_value, read_is_plain = self.get_config(val[3:])
                except KeyError as e:
                    exc = click.UsageError(
                        "Could not find configuration '%s' in INFO file" % val
                    )
                    if click_obj:
                        click_obj.delayed_config_exception = exc
                        return None
                    raise exc from e
                if read_is_plain != is_plain:
                    raise click.UsageError(
                        "Configuration '%s' mismatched `plain` attribute -- "
                        "this is a bug, please report it." % val[3:]
                    )
                flow_cls._flow_state.self_data[FlowStateItems.CONFIGS][name] = (
                    read_value,
                    is_plain,
                )
                to_return[name] = (
                    read_value
                    if read_value is None or is_plain
                    else ConfigValue(read_value)
                )
            else:
                if parser:
                    read_value = self._call_parser(parser, val, is_plain)
                else:
                    if is_plain:
                        read_value = val
                    else:
                        try:
                            read_value = json.loads(val)
                        except json.JSONDecodeError as e:
                            msgs.append(
                                "configuration value for '%s' is not valid JSON: %s"
                                % (name, e)
                            )
                            continue
                        # TODO: Support YAML
                flow_cls._flow_state.self_data[FlowStateItems.CONFIGS][name] = (
                    read_value,
                    is_plain,
                )
                to_return[name] = (
                    read_value
                    if read_value is None or is_plain
                    else ConfigValue(read_value)
                )

        reqs = missing_configs.intersection(self._req_configs)
        for missing in reqs:
            msgs.append("missing configuration for '%s'" % missing)
        for missing in no_file:
            msgs.append(
                "configuration file '%s' could not be read for '%s'"
                % (merged_configs[missing][len(_CONVERTED_NO_FILE) :], missing)
            )
        for missing in no_default_file:
            msgs.append(
                "default configuration file '%s' could not be read for '%s'"
                % (merged_configs[missing][len(_CONVERTED_DEFAULT_NO_FILE) :], missing)
            )
        if msgs:
            exc = click.UsageError(
                "Bad values passed for configuration options: %s" % ", ".join(msgs)
            )
            if click_obj:
                click_obj.delayed_config_exception = exc
                return None
            raise exc

        debug.userconf_exec("Finalized configs: %s" % str(to_return))
        return to_return

    def process_configs_click(self, ctx, param, value):
        return self.process_configs(
            ctx.obj.flow.name,
            param.name,
            dict(value),
            ctx.params["quiet"],
            ctx.params["datastore"],
            click_obj=ctx.obj,
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "ConfigInput"

    @staticmethod
    def _call_parser(parser, val, is_plain):
        if isinstance(parser, str):
            if len(parser) and parser[0] == ".":
                parser = "metaflow" + parser
            path, func = parser.rsplit(".", 1)
            try:
                func_module = importlib.import_module(path)
            except ImportError as e:
                raise ValueError("Cannot locate parser %s" % parser) from e
            parser = getattr(func_module, func, None)
            if parser is None or not callable(parser):
                raise ValueError(
                    "Parser %s is either not part of %s or not a callable"
                    % (func, path)
                )
        return_value = parser(val)
        if not is_plain and not isinstance(return_value, Mapping):
            raise ValueError(
                "Parser %s returned a value that is not a mapping (got type %s): %s"
                % (str(parser), type(return_value), return_value)
            )
        return return_value


class LocalFileInput(click.Path):
    # Small wrapper around click.Path to set the value from which to read configuration
    # values. This is set immediately upon processing the --local-config-file
    # option and will therefore then be available when processing any of the other
    # --config options (which will call ConfigInput.process_configs)
    name = "LocalFileInput"

    def convert(self, value, param, ctx):
        v = super().convert(value, param, ctx)
        ConfigInput.set_config_file(value)
        return v

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "LocalFileInput"


def config_options_with_config_input(cmd):
    help_strs = []
    required_names = []
    defaults = {}
    config_seen = set()
    parsers = {}
    flow_cls = getattr(current_flow, "flow_cls", None)
    if flow_cls is None:
        return cmd, None

    parameters = [p for _, p in flow_cls._get_parameters() if p.IS_CONFIG_PARAMETER]
    # List all the configuration options
    for arg in parameters[::-1]:
        kwargs = arg.option_kwargs(False)
        if arg.name in config_seen:
            msg = (
                "Multiple configurations use the same name '%s'. Please change the "
                "names of some of your configurations" % arg.name
            )
            raise MetaflowException(msg)
        config_seen.add(arg.name)
        if kwargs["required"]:
            required_names.append(arg.name)

        defaults[arg.name] = (
            arg.kwargs.get("default", None),
            arg._default_is_file,
        )
        help_strs.append("  - %s: %s" % (arg.name, kwargs.get("help", "")))
        parsers[arg.name] = (arg.parser, arg.kwargs["plain"])

    if not config_seen:
        # No configurations -- don't add anything; we set it to False so that it
        # can be checked whether or not we called this.
        return cmd, False

    help_str = (
        "Configuration options for the flow. "
        "Multiple configurations can be specified. Cannot be used with resume."
    )
    help_str = "\n\n".join([help_str] + help_strs)
    config_input = ConfigInput(required_names, defaults, parsers)
    cb_func = config_input.process_configs_click

    cmd.params.insert(
        0,
        click.Option(
            ["--config-value", "config_value"],
            nargs=2,
            multiple=True,
            type=MultipleTuple([click.Choice(config_seen), ConvertDictOrStr()]),
            callback=cb_func,
            help=help_str,
            envvar="METAFLOW_FLOW_CONFIG_VALUE",
            show_default=False,
            default=[
                (
                    k,
                    (
                        ConvertDictOrStr.mark_as_default(v[0])
                        if not callable(v[0]) and not v[1]
                        else None
                    ),
                )
                for k, v in defaults.items()
            ],
            required=False,
        ),
    )
    cmd.params.insert(
        0,
        click.Option(
            ["--config", "config"],
            nargs=2,
            multiple=True,
            type=MultipleTuple([click.Choice(config_seen), ConvertPath()]),
            callback=cb_func,
            help=help_str,
            envvar="METAFLOW_FLOW_CONFIG",
            show_default=False,
            default=[
                (
                    k,
                    (
                        ConvertPath.mark_as_default(v[0])
                        if not callable(v[0]) and v[1]
                        else None
                    ),
                )
                for k, v in defaults.items()
            ],
            required=False,
        ),
    )
    return cmd, config_input


def config_options(cmd):
    cmd, _ = config_options_with_config_input(cmd)
    return cmd



class AzureDefaultClientProvider(object):
    name = "azure-default"

    @staticmethod
    def create_cacheable_azure_credential(*args, **kwargs):
        """azure.identity.DefaultAzureCredential is not readily cacheable in a dictionary
        because it does not have a content based hash and equality implementations.

        We implement a subclass CacheableDefaultAzureCredential to add them.

        We need this because credentials will be part of the cache key in _ClientCache.
        """
        from azure.identity import DefaultAzureCredential

        class CacheableDefaultAzureCredential(DefaultAzureCredential):
            def __init__(self, *args, **kwargs):
                super(CacheableDefaultAzureCredential, self).__init__(*args, **kwargs)
                # Just hashing all the kwargs works because they are all individually
                # hashable as of 7/15/2022.
                #
                # What if Azure adds unhashable things to kwargs?
                # - We will have CI to catch this (it will always install the latest Azure SDKs)
                # - In Metaflow usage today we never specify any kwargs anyway. (see last line
                #   of the outer function.
                self._hash_code = hash((args, tuple(sorted(kwargs.items()))))

            def __hash__(self):
                return self._hash_code

            def __eq__(self, other):
                return hash(self) == hash(other)

        return CacheableDefaultAzureCredential(*args, **kwargs)


cached_provider_class = None


def create_cacheable_azure_credential():
    global cached_provider_class
    if cached_provider_class is None:
        from metaflow.metaflow_config import DEFAULT_AZURE_CLIENT_PROVIDER
        from metaflow.plugins import AZURE_CLIENT_PROVIDERS

        for p in AZURE_CLIENT_PROVIDERS:
            if p.name == DEFAULT_AZURE_CLIENT_PROVIDER:
                cached_provider_class = p
                break
        else:
            raise ValueError(
                "Cannot find Azure Client provider %s" % DEFAULT_AZURE_CLIENT_PROVIDER
            )
    return cached_provider_class.create_cacheable_azure_credential()



from collections import namedtuple
import functools
import re

from metaflow.metaflow_config import NAMESPACED_EVENTS_PREFIX
from metaflow.util import is_stringish

ParsedEvent = namedtuple(
    "ParsedEvent",
    ["full_name", "project", "branch", "logical_name", "is_namespaced"],
)


def namespaced_event_name(event_name):
    """
    Creates a project-namespaced event name based on @project settings.

    Use this to automatically prefix event names with your @project
    namespace, ensuring events are isolated per project and branch.

    The resolved name follows the format:
        mfns.<project>.<branch>.<event_name>

    If no @project decorator is present, resolves to:
        mfns.<event_name>

    Parameters
    ----------
    event_name : str
        The logical event name (e.g., 'data_ready')

    Returns
    -------
    Callable[..., str]
        A callable that returns the fully namespaced event name (str) when invoked

    Examples
    --------
    With @trigger decorator:

        ```
        from metaflow import namespaced_event_name, project

        @project(name="foo")
        @trigger(event=namespaced_event_name('data_ready'))
        class MyFlow(FlowSpec):
            ...

        @project(name="foo")
        @trigger(event={'name': namespaced_event_name('data_ready'), 'parameters': {'x': 'y'}})
        class MyFlow(FlowSpec):
        ```

    With ArgoEvent:

        ```
        from metaflow import namespaced_event_name
        from metaflow.integrations import ArgoEvent

        ArgoEvent(namespaced_event_name('data_ready')).publish()
        ```
    """
    if event_name is None or not is_stringish(event_name):
        raise ValueError(
            "event_name should be a string not %s" % (str(type(event_name)))
        )

    def _delayed_name_generator(context=None, event_name=None):
        from metaflow import current

        project_name = current.get("project_name", None)
        branch_name = current.get("branch_name", None)
        return _make_namespaced_event_name(event_name, project_name, branch_name)

    return functools.partial(_delayed_name_generator, event_name=event_name)


def _make_namespaced_event_name(logical_name, project=None, branch=None):
    """
    Construct a namespaced event name from components.

    Parameters
    ----------
    logical_name : str
        The base event name
    project : str, optional
        Project name from @project decorator
    branch : str, optional
        Branch name from @project decorator

    Returns
    -------
    str
        Namespaced event name in format mfns.<project>.<branch>.<logical_name>
        or mfns.<logical_name> if no project/branch
    """
    if project and branch:
        # When it comes to project and branches we want to make
        # sure that we are not having any non-acceptable characters.
        # allow dot (.) in project and branch names
        branch = re.sub(r"[^a-zA-Z0-9_\-\.]", "", branch)
        project = re.sub(r"[^a-zA-Z0-9_\-\.]", "", project)
        return f"{NAMESPACED_EVENTS_PREFIX}.{project}.{branch}.{logical_name}"
    else:
        return f"{NAMESPACED_EVENTS_PREFIX}.{logical_name}"



import itertools
import json
from abc import ABC, abstractmethod

from .. import metaflow_config

from .content_addressed_store import ContentAddressedStore
from .task_datastore import TaskDataStore
from .spin_datastore import SpinTaskDatastore
from ..metaflow_profile import from_start


class FlowDataStore(object):
    default_storage_impl = None

    def __init__(
        self,
        flow_name,
        environment=None,
        metadata=None,
        event_logger=None,
        monitor=None,
        storage_impl=None,
        ds_root=None,
    ):
        """
        Initialize a Flow level datastore.

        This datastore can then be used to get TaskDataStore to store artifacts
        and metadata about a task as well as a ContentAddressedStore to store
        things like packages, etc.

        Parameters
        ----------
        flow_name : str
            The name of the flow
        environment : MetaflowEnvironment, optional
            Environment this datastore is operating in
        metadata : MetadataProvider, optional
            The metadata provider to use and update if needed, by default None
        event_logger : EventLogger, optional
            EventLogger to use to report events, by default None
        monitor : Monitor, optional
            Monitor to use to measure/monitor events, by default None
        storage_impl : type
            Class for the backing DataStoreStorage to use; if not provided use
            default_storage_impl, optional
        ds_root : str
            The optional root for this datastore; if not provided, use the
            default for the DataStoreStorage, optional
        """
        storage_impl = storage_impl if storage_impl else self.default_storage_impl
        if storage_impl is None:
            raise RuntimeError("No datastore storage implementation specified")
        self._storage_impl = storage_impl(ds_root)
        self.TYPE = self._storage_impl.TYPE

        # Public attributes
        self.flow_name = flow_name
        self.environment = environment
        self.metadata = metadata
        self.logger = event_logger
        self.monitor = monitor

        self.ca_store = ContentAddressedStore(
            self._storage_impl.path_join(self.flow_name, "data"), self._storage_impl
        )

        # Private
        self._metadata_cache = None

    @property
    def datastore_root(self):
        return self._storage_impl.datastore_root

    def set_metadata_cache(self, cache):
        self._metadata_cache = cache

    def get_task_datastores(
        self,
        run_id=None,
        steps=None,
        pathspecs=None,
        allow_not_done=False,
        attempt=None,
        include_prior=False,
        mode="r",
        join_type=None,
        orig_flow_datastore=None,
        spin_artifacts=None,
    ):
        """
        Return a list of TaskDataStore for a subset of the tasks.

        We filter the list based on `steps` if non-None.
        Alternatively, `pathspecs` can contain the exact list of pathspec(s)
        (run_id/step_name/task_id) that should be filtered.
        Note: When `pathspecs` is specified, we expect strict consistency and
        not eventual consistency in contrast to other modes.

        Parameters
        ----------
        run_id : str, optional
            Run ID to get the tasks from. If not specified, use pathspecs,
            by default None
        steps : List[str] , optional
            Steps to get the tasks from. If run_id is specified, this
            must also be specified, by default None
        pathspecs : List[str], optional
            Full task specs (run_id/step_name/task_id[/attempt]). Can be used instead of
            specifying run_id and steps, by default None
        allow_not_done : bool, optional
            If True, returns the latest attempt of a task even if that attempt
            wasn't marked as done, by default False
        attempt : int, optional
            Attempt number of the tasks to return.  If not provided, returns latest attempt.
        include_prior : boolean, default False
            If True, returns all attempts up to and including attempt.
        mode : str, default "r"
            Mode to initialize the returned TaskDataStores in.
        join_type : str, optional, default None
            If specified, the join type for the task. This is used to determine
            the user specified artifacts for the task in case of a spin task.
        orig_flow_datastore : MetadataProvider, optional, default None
            The metadata provider in case of a spin task. If provided, the
            returned TaskDataStore will be a SpinTaskDatastore instead of a
            TaskDataStore.
        spin_artifacts : Dict[str, Any], optional, default None
            Artifacts provided by user that can override the artifacts fetched via the
            spin pathspec.

        Returns
        -------
        List[TaskDataStore]
            Task datastores for all the tasks specified.
        """
        task_urls = []
        # Note: When `pathspecs` is specified, we avoid the potentially
        # eventually consistent `list_content` operation, and directly construct
        # the task_urls list.
        if pathspecs:
            task_urls = [
                self._storage_impl.path_join(self.flow_name, pathspec)
                for pathspec in pathspecs
            ]
        else:
            run_prefix = self._storage_impl.path_join(self.flow_name, run_id)
            if steps:
                step_urls = [
                    self._storage_impl.path_join(run_prefix, step) for step in steps
                ]
            else:
                step_urls = [
                    step.path
                    for step in self._storage_impl.list_content([run_prefix])
                    if step.is_file is False
                ]
            task_urls = [
                task.path
                for task in self._storage_impl.list_content(step_urls)
                if task.is_file is False
            ]
        urls = []
        # parse content urls for specific attempt only, or for all attempts in max range
        attempt_range = range(metaflow_config.MAX_ATTEMPTS)
        # we have no reason to check for attempts greater than MAX_ATTEMPTS, as they do not exist.
        if attempt is not None and attempt <= metaflow_config.MAX_ATTEMPTS - 1:
            attempt_range = range(attempt + 1) if include_prior else [attempt]
        for task_url in task_urls:
            # task_url can have a trailing slash, so strip this to avoid empty strings in the split
            task_splits = task_url.rstrip("/").split("/")
            # Usually it is flow, run, step, task (so 4 components) -- if we have a
            # fifth one, there is a specific attempt number listed as well.
            task_attempt_range = attempt_range
            if len(task_splits) == 5:
                task_attempt_range = [int(task_splits[4])]
            for attempt in task_attempt_range:
                for suffix in [
                    TaskDataStore.METADATA_DATA_SUFFIX,
                    TaskDataStore.METADATA_ATTEMPT_SUFFIX,
                    TaskDataStore.METADATA_DONE_SUFFIX,
                ]:
                    urls.append(
                        self._storage_impl.path_join(
                            task_url,
                            TaskDataStore.metadata_name_for_attempt(suffix, attempt),
                        )
                    )

        latest_started_attempts = {}
        done_attempts = set()
        data_objs = {}
        with self._storage_impl.load_bytes(urls) as get_results:
            for key, path, meta in get_results:
                if path is not None:
                    _, run, step, task, fname = self._storage_impl.path_split(key)
                    attempt, fname = TaskDataStore.parse_attempt_metadata(fname)
                    attempt = int(attempt)
                    if fname == TaskDataStore.METADATA_DONE_SUFFIX:
                        done_attempts.add((run, step, task, attempt))
                    elif fname == TaskDataStore.METADATA_ATTEMPT_SUFFIX:
                        latest_started_attempts[(run, step, task)] = max(
                            latest_started_attempts.get((run, step, task), 0), attempt
                        )
                    elif fname == TaskDataStore.METADATA_DATA_SUFFIX:
                        # This somewhat breaks the abstraction since we are using
                        # load_bytes directly instead of load_metadata
                        with open(path, encoding="utf-8") as f:
                            data_objs[(run, step, task, attempt)] = json.load(f)
        # We now figure out the latest attempt that started *and* finished.
        # Note that if an attempt started but didn't finish, we do *NOT* return
        # the previous attempt
        latest_started_attempts = set(
            (run, step, task, attempt)
            for (run, step, task), attempt in latest_started_attempts.items()
        )
        if allow_not_done:
            latest_to_fetch = (
                done_attempts.union(latest_started_attempts)
                if include_prior
                else latest_started_attempts
            )
        else:
            latest_to_fetch = (
                done_attempts
                if include_prior
                else (latest_started_attempts & done_attempts)
            )
        latest_to_fetch = [
            (
                v[0],
                v[1],
                v[2],
                v[3],
                data_objs.get(v),
                mode,
                allow_not_done,
                join_type,
                orig_flow_datastore,
                spin_artifacts,
            )
            for v in latest_to_fetch
        ]
        return list(itertools.starmap(self.get_task_datastore, latest_to_fetch))

    def get_task_datastore(
        self,
        run_id,
        step_name,
        task_id,
        attempt=None,
        data_metadata=None,
        mode="r",
        allow_not_done=False,
        join_type=None,
        orig_flow_datastore=None,
        spin_artifacts=None,
        persist=True,
    ):
        if orig_flow_datastore is not None:
            # In spin step subprocess, use SpinTaskDatastore for accessing artifacts
            if join_type is not None:
                # If join_type is specified, we need to use the artifacts corresponding
                # to that particular join index, specified by the parent task pathspec.
                spin_artifacts = spin_artifacts.get(
                    f"{run_id}/{step_name}/{task_id}", {}
                )
            from_start(
                "FlowDataStore: get_task_datastore for spin task for type %s %s metadata"
                % (self.TYPE, "without" if data_metadata is None else "with")
            )
            # Get the task datastore for the spun task.
            orig_datastore = orig_flow_datastore.get_task_datastore(
                run_id,
                step_name,
                task_id,
                attempt=attempt,
                data_metadata=data_metadata,
                mode=mode,
                allow_not_done=allow_not_done,
                persist=persist,
            )

            return SpinTaskDatastore(
                self.flow_name,
                run_id,
                step_name,
                task_id,
                orig_datastore,
                spin_artifacts,
            )

        cache_hit = False
        if (
            self._metadata_cache is not None
            and data_metadata is None
            and attempt is not None
            and allow_not_done is False
        ):
            # If we have a metadata cache, we can try to load the metadata
            # from the cache if it is not provided.
            data_metadata = self._metadata_cache.load_metadata(
                run_id, step_name, task_id, attempt
            )
            cache_hit = data_metadata is not None

        from_start(
            "FlowDataStore: get_task_datastore for regular task for type %s %s metadata"
            % (self.TYPE, "without" if data_metadata is None else "with")
        )
        task_datastore = TaskDataStore(
            self,
            run_id,
            step_name,
            task_id,
            attempt=attempt,
            data_metadata=data_metadata,
            mode=mode,
            allow_not_done=allow_not_done,
            persist=persist,
        )

        # Only persist in cache if it is non-changing (so done only) and we have
        # a non-None attempt
        if (
            not cache_hit
            and self._metadata_cache is not None
            and allow_not_done is False
            and attempt is not None
        ):
            self._metadata_cache.store_metadata(
                run_id, step_name, task_id, attempt, task_datastore.ds_metadata
            )

        return task_datastore

    def save_data(self, data_iter, len_hint=0):
        """Saves data to the underlying content-addressed store

        Parameters
        ----------
        data_iter : Iterator[bytes]
            Iterator over blobs to save; each item in the list will be saved individually.
        len_hint : int
            Estimate of the number of items that will be produced by the iterator,
            by default 0.

        Returns
        -------
        (str, str)
            Tuple containing the URI to access the saved resource as well as
            the key needed to retrieve it using load_data. This is returned in
            the same order as the input.
        """
        save_results = self.ca_store.save_blobs(data_iter, raw=True, len_hint=len_hint)
        return [(r.uri, r.key) for r in save_results]

    def load_data(self, keys, force_raw=False):
        """Retrieves data from the underlying content-addressed store

        Parameters
        ----------
        keys : List[str]
            Keys to retrieve
        force_raw : bool, optional
            Backward compatible mode. Raw data will be properly identified with
            metadata information but older datastores did not do this. If you
            know the data should be handled as raw data, set this to True,
            by default False

        Returns
        -------
        Iterator[bytes]
            Iterator over (key, blob) tuples
        """
        for key, blob in self.ca_store.load_blobs(keys, force_raw=force_raw):
            yield key, blob


class MetadataCache(ABC):
    @abstractmethod
    def load_metadata(self, run_id, step_name, task_id, attempt):
        raise NotImplementedError()

    @abstractmethod
    def store_metadata(self, run_id, step_name, task_id, attempt, metadata_dict):
        raise NotImplementedError()



import importlib
import itertools
import pickle
import socket
import sys
import traceback

from . import data_transferer

from .consts import (
    FIELD_ARGS,
    FIELD_CONTENT,
    FIELD_KWARGS,
    FIELD_MSGTYPE,
    FIELD_OPTYPE,
    FIELD_TARGET,
    MSG_CONTROL,
    MSG_EXCEPTION,
    MSG_INTERNAL_ERROR,
    MSG_OP,
    MSG_REPLY,
    OP_GETATTR,
    OP_SETATTR,
    OP_DELATTR,
    OP_CALL,
    OP_CALLATTR,
    OP_REPR,
    OP_STR,
    OP_HASH,
    OP_PICKLE,
    OP_DEL,
    OP_GETMETHODS,
    OP_DIR,
    OP_CALLFUNC,
    OP_CALLONCLASS,
    OP_GETVAL,
    OP_SETVAL,
    OP_INIT,
    OP_SUBCLASSCHECK,
    OP_GETCLASSATTR,
    VALUE_LOCAL,
    VALUE_REMOTE,
    CONTROL_GETEXPORTS,
    CONTROL_SHUTDOWN,
)

from .communication.channel import Channel
from .communication.socket_bytestream import SocketByteStream
from .communication.utils import __try_op__

from .data_transferer import DataTransferer, ObjReference
from .override_decorators import (
    RemoteAttrOverride,
    RemoteOverride,
    RemoteExceptionSerializer,
)
from .exception_transferer import dump_exception
from .utils import get_methods, get_canonical_name

BIND_TIMEOUT = 0.1
BIND_RETRY = 1


class Server(object):
    def __init__(self, config_dir, max_pickle_version):
        self._max_pickle_version = data_transferer.defaultProtocol = max_pickle_version
        try:
            mappings = importlib.import_module(".server_mappings", package=config_dir)
        except Exception as e:
            raise RuntimeError(
                "Cannot import server_mappings from '%s': %s" % (sys.path[0], str(e))
            )
        try:
            # Import module as a relative package to make sure that it is consistent
            # with how the client does it -- this enables us to do the same type of
            # relative imports in overrides
            override_module = importlib.import_module(".overrides", package=config_dir)
            override_values = override_module.__dict__.values()
        except ImportError:
            # We ignore so the file can be non-existent if not needed
            override_values = []
        except Exception as e:
            raise RuntimeError(
                "Cannot import overrides from '%s': %s" % (sys.path[0], str(e))
            )

        self._aliases = {}
        self._known_classes, a1 = self._flatten_dict(mappings.EXPORTED_CLASSES)
        self._class_types_to_names = {v: k for k, v in self._known_classes.items()}
        self._known_funcs, a2 = self._flatten_dict(mappings.EXPORTED_FUNCTIONS)
        self._known_vals, a3 = self._flatten_dict(mappings.EXPORTED_VALUES)
        self._known_exceptions, a4 = self._flatten_dict(mappings.EXPORTED_EXCEPTIONS)
        self._proxied_types = {
            "%s.%s" % (t.__module__, t.__name__): t for t in mappings.PROXIED_CLASSES
        }
        self._class_types_to_names.update(
            {v: k for k, v in self._proxied_types.items()}
        )

        # We will also proxy functions and methods from objects as needed. This is useful
        # for defaultdict for example since the `default_factory` function is a
        # lambda that needs to be transferred. Methods are also proxied to support
        # bound methods returned from functions (e.g., evaluator patterns).
        self._class_types_to_names[type(lambda x: x)] = "function"

        # Register method type for bound methods
        class _TempClass:
            def _temp_method(self):
                pass

        self._class_types_to_names[type(_TempClass()._temp_method)] = "method"

        # Update all alias information
        for base_name, aliases in itertools.chain(
            a1.items(), a2.items(), a3.items(), a4.items()
        ):
            for alias in aliases:
                a = self._aliases.setdefault(alias, base_name)
                if a != base_name:
                    # Technically we could have a that aliases b and b that aliases c
                    # and then a that aliases c. This would error out in that case
                    # even though it is valid. It is easy for the user to get around
                    # this by listing aliases in the same order so we don't support
                    # it for now.
                    raise ValueError(
                        "%s is an alias to both %s and %s -- make sure all aliases "
                        "are listed in the same order" % (alias, base_name, a)
                    )

        # Detect circular aliases. If a user lists ("a", "b") and then ("b", "a"), we
        # will have an entry in aliases saying b is an alias for a and a is an alias
        # for b which is a recipe for disaster since we no longer have a cannonical name
        # for things.
        for alias, base_name in self._aliases.items():
            if base_name in self._aliases:
                raise ValueError(
                    "%s and %s are circular aliases -- make sure all aliases "
                    "are listed in the same order" % (alias, base_name)
                )

        # Determine if we have any overrides
        self._overrides = {}
        self._getattr_overrides = {}
        self._setattr_overrides = {}
        self._exception_serializers = {}
        for override in override_values:
            if isinstance(override, (RemoteAttrOverride, RemoteOverride)):
                for obj_name, obj_funcs in override.obj_mapping.items():
                    canonical_name = get_canonical_name(obj_name, self._aliases)
                    obj_type = self._known_classes.get(
                        canonical_name, self._proxied_types.get(obj_name)
                    )
                    if obj_type is None:
                        raise ValueError(
                            "%s does not refer to a proxied or exported type" % obj_name
                        )
                    if isinstance(override, RemoteOverride):
                        override_dict = self._overrides.setdefault(obj_type, {})
                    elif override.is_setattr:
                        override_dict = self._setattr_overrides.setdefault(obj_type, {})
                    else:
                        override_dict = self._getattr_overrides.setdefault(obj_type, {})
                    if isinstance(obj_funcs, str):
                        obj_funcs = (obj_funcs,)
                    for name in obj_funcs:
                        if name in override_dict:
                            raise ValueError(
                                "%s was already overridden for %s" % (name, obj_name)
                            )
                        override_dict[name] = override.func
            elif isinstance(override, RemoteExceptionSerializer):
                canonical_name = get_canonical_name(override.class_path, self._aliases)
                if canonical_name not in self._known_exceptions:
                    raise ValueError(
                        "%s does not refer to an exported exception"
                        % override.class_path
                    )
                if override.class_path in self._exception_serializers:
                    raise ValueError(
                        "%s exception serializer already defined" % override.class_path
                    )
                self._exception_serializers[canonical_name] = override.serializer

        # Process the exceptions making sure we have all the ones we need and building a
        # topologically sorted list for the client to instantiate
        name_to_parent_count = {}
        name_to_parents = {}
        parent_to_child = {}

        for ex_name, ex_cls in self._known_exceptions.items():
            ex_name_canonical = get_canonical_name(ex_name, self._aliases)
            parents = []
            for base in ex_cls.__mro__[1:]:
                if base is object:
                    raise ValueError(
                        "Exported exceptions not rooted in a builtin exception "
                        "are not supported: %s." % ex_name
                    )
                if base.__module__ == "builtins":
                    # We found our base exception
                    parents.append("builtins." + base.__name__)
                    break
                else:
                    fqn = ".".join([base.__module__, base.__name__])
                    canonical_fqn = get_canonical_name(fqn, self._aliases)
                    if canonical_fqn in self._known_exceptions:
                        parents.append(canonical_fqn)
                        children = parent_to_child.setdefault(canonical_fqn, [])
                        children.append(ex_name_canonical)
                    else:
                        raise ValueError(
                            "Exported exception %s has non exported and non builtin parent "
                            "exception: %s (%s). Known exceptions: %s."
                            % (ex_name, fqn, canonical_fqn, str(self._known_exceptions))
                        )
            name_to_parent_count[ex_name_canonical] = len(parents) - 1
            name_to_parents[ex_name_canonical] = parents

        # We now form the exceptions and put them in self._known_exceptions in
        # the proper order (topologically)
        self._known_exceptions = []
        # Find roots
        to_process = []
        for name, count in name_to_parent_count.items():
            if count == 0:
                to_process.append(name)

        # Topologically process the exceptions
        while to_process:
            next_round = []
            for name in to_process:
                self._known_exceptions.append((name, name_to_parents[name]))
                del name_to_parent_count[name]
                for child in parent_to_child.get(name, []):
                    cur_child_count = name_to_parent_count[child]
                    if cur_child_count == 1:
                        next_round.append(child)
                    else:
                        name_to_parent_count[child] = cur_child_count - 1
            to_process = next_round

        if name_to_parent_count:
            raise ValueError(
                "Badly rooted exceptions: %s" % ", ".join(name_to_parent_count.keys())
            )
        self._active = False
        self._channel = None
        self._datatransferer = DataTransferer(self)

        self._handlers = {
            OP_GETATTR: self._handle_getattr,
            OP_SETATTR: self._handle_setattr,
            OP_DELATTR: self._handle_delattr,
            OP_CALL: self._handle_call,
            OP_CALLATTR: self._handle_callattr,
            OP_REPR: self._handle_repr,
            OP_STR: self._handle_str,
            OP_HASH: self._handle_hash,
            OP_PICKLE: self._handle_pickle,
            OP_DEL: self._handle_del,
            OP_GETMETHODS: self._handle_getmethods,
            OP_DIR: self._handle_dir,
            OP_CALLFUNC: self._handle_callfunc,
            OP_CALLONCLASS: self._handle_callonclass,
            OP_GETVAL: self._handle_getval,
            OP_SETVAL: self._handle_setval,
            OP_INIT: self._handle_init,
            OP_SUBCLASSCHECK: self._handle_subclasscheck,
            OP_GETCLASSATTR: self._handle_getclassattr,
        }

        self._local_objects = {}

    def serve(self, path=None, port=None):
        # Open up a connection
        if path is not None:
            # Keep the print line to facilitate debugging
            # print("SERVER: Starting at %s" % path)
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            __try_op__("bind", sock.bind, BIND_RETRY, path)
        elif port is not None:
            family, socktype, proto, _, sockaddr = socket.getaddrinfo(
                "127.0.0.1", port, socket.AF_INET, socket.SOCK_STREAM
            )
            sock = socket.socket(family=family, type=socktype)
            __try_op__("bind", sock.bind, BIND_RETRY, sockaddr)

        # We assume -- and this is true for this use case -- that we will have
        # one and only one "client". The client will connect once and we assume
        # that if the connection drops, we lost the client and we can die.
        sock.listen(0)
        self._active = True
        connection, remote = sock.accept()
        self._channel = Channel(SocketByteStream(connection))
        while self._active:
            # This will block until we get a full message
            self._dispatch_request(self._channel.recv())

    def encode(self, obj):
        # This encodes an object to transfer back out
        # Basic data types will be sent over directly
        # and other ones will be saved locally and an ID will be returned
        return self._datatransferer.dump(obj)

    def encode_exception(self, ex_type, ex, trace_back):
        try:
            full_name = "%s.%s" % (ex_type.__module__, ex_type.__name__)
            get_canonical_name(full_name, self._aliases)
            serializer = self._exception_serializers.get(full_name)
        except AttributeError:
            # Ignore if no __module__ for example -- definitely not something we built
            serializer = None
        extra_content = None
        if serializer is not None:
            extra_content = serializer(ex)
        return dump_exception(
            self._datatransferer, ex_type, ex, trace_back, extra_content
        )

    def decode(self, json_obj):
        # This decodes an object that was transferred in
        return self._datatransferer.load(json_obj)

    def _dispatch_request(self, json_request):
        try:
            if json_request[FIELD_MSGTYPE] == MSG_CONTROL:
                if json_request[FIELD_OPTYPE] == CONTROL_SHUTDOWN:
                    self._active = False
                    self._reply(MSG_REPLY, self.encode(None))
                    return
                elif json_request[FIELD_OPTYPE] == CONTROL_GETEXPORTS:
                    self._reply(
                        MSG_REPLY,
                        {
                            "classes": list(self._known_classes),
                            "functions": list(self._known_funcs),
                            "values": list(self._known_vals),
                            "exceptions": self._known_exceptions,
                            "proxied": list(self._proxied_types),
                            "aliases": self._aliases,
                        },
                    )
                    return
                else:
                    raise ValueError("Unknown control message")
            if json_request[FIELD_MSGTYPE] != MSG_OP:
                raise ValueError("Invalid message received: %s" % json_request)
            op_type = json_request[FIELD_OPTYPE]
            op_target = json_request.get(FIELD_TARGET)
            if op_target is not None:
                op_target = self.decode(op_target)
            args = json_request.get(FIELD_ARGS)
            if args is not None:
                args = self.decode(args)
            kwargs = json_request.get(FIELD_KWARGS)
            if kwargs is not None:
                kwargs = dict(self.decode(kwargs))

            result = self._handlers[op_type](op_target, *args, **kwargs)
        except:  # noqa E722
            ex_type, ex, trace_back = sys.exc_info()
            if ex_type is SystemExit or ex_type is KeyboardInterrupt:
                raise
            try:
                self._reply(
                    MSG_EXCEPTION, self.encode_exception(ex_type, ex, trace_back)
                )
            except (SystemExit, KeyboardInterrupt):
                raise
            except:  # noqa E722
                internal_err = traceback.format_exc()
                self._reply(MSG_INTERNAL_ERROR, internal_err)
        else:
            try:
                self._reply(MSG_REPLY, self.encode(result))
            except (SystemExit, KeyboardInterrupt):
                raise
            except:  # noqa E722
                internal_err = traceback.format_exc()
                self._reply(MSG_INTERNAL_ERROR, internal_err)

    def _reply(self, reply_type, content):
        sys.stdout.flush()
        sys.stderr.flush()
        self._channel.send({FIELD_MSGTYPE: reply_type, FIELD_CONTENT: content})

    def can_pickle(self, obj):
        return self._class_types_to_names.get(type(obj)) is not None

    def pickle_object(self, obj):
        # This function is called to pickle obj to be transferred to the client
        # when the data layer can't transfer it by itself. We basically will save
        # locally and transfer an identifier for it
        identifier = id(obj)
        mapped_class_name = self._class_types_to_names.get(type(obj))
        if mapped_class_name is None:
            raise ValueError("Cannot proxy value of type %s" % type(obj))
        self._local_objects[identifier] = obj
        return ObjReference(VALUE_REMOTE, mapped_class_name, identifier)

    def unpickle_object(self, obj):
        if (not isinstance(obj, ObjReference)) or obj.value_type != VALUE_LOCAL:
            raise ValueError("Invalid transferred object: %s" % str(obj))
        result = self._local_objects.get(obj.identifier)
        if result is not None:
            return result
        raise ValueError("Invalid object -- id %s not known" % obj.identifier)

    @staticmethod
    def _flatten_dict(d):
        # Takes a dictionary of ("name1", "name2"): {"sub1": X, "sub2": Y} and
        # returns one of "name1.sub1": X, "name1.sub2": Y, etc. as well as a
        # dictionary of aliases {"name1": ["name2"]}...
        result = {}
        aliases = {}
        for base, values in d.items():
            if isinstance(base, tuple):
                aliases[base[0]] = list(base[1:])
                base = base[0]
            for name, value in values.items():
                result["%s.%s" % (base, name)] = value
        return result, aliases

    def _handle_getattr(self, target, name):
        override_mapping = self._getattr_overrides.get(type(target))
        if override_mapping:
            override_func = override_mapping.get(name)
            if override_func:
                return override_func(target, name)
        return getattr(target, name)

    def _handle_setattr(self, target, name, value):
        override_mapping = self._setattr_overrides.get(type(target))
        if override_mapping:
            override_func = override_mapping.get(name)
            if override_func:
                return override_func(target, name, value)
        return setattr(target, name, value)

    def _handle_delattr(self, target, name):
        delattr(target, name)

    def _handle_call(self, target, *args, **kwargs):
        return self._handle_callattr(target, "__call__", *args, **kwargs)

    def _handle_callattr(self, target, name, *args, **kwargs):
        attr = getattr(target, name)
        override_mapping = self._overrides.get(type(target))
        if override_mapping:
            override_func = override_mapping.get(name)
            if override_func:
                return override_func(target, attr, *args, **kwargs)
        return attr(*args, **kwargs)

    def _handle_repr(self, target):
        return repr(target)

    def _handle_str(self, target):
        return str(target)

    def _handle_hash(self, target):
        return hash(target)

    def _handle_pickle(self, target, proto):
        effective_protocol = min(self._max_pickle_version, proto)
        return pickle.dumps(target, protocol=effective_protocol)

    def _handle_del(self, target):
        del target

    def _handle_getmethods(self, target, class_name):
        class_type = self._known_classes.get(class_name)
        if class_type is None:
            class_type = self._proxied_types.get(class_name)
        if class_type is None:
            raise ValueError("Unknown class %s" % class_name)
        return get_methods(class_type)

    def _handle_dir(self, target):
        return dir(target)

    def _handle_callfunc(self, target, name, *args, **kwargs):
        func_to_call = self._known_funcs.get(name)
        if func_to_call is None:
            raise ValueError("Unknown function %s" % name)
        return func_to_call(*args, **kwargs)

    def _handle_callonclass(self, target, class_name, name, is_static, *args, **kwargs):
        class_type = self._known_classes.get(class_name)
        if class_type is None:
            raise ValueError("Unknown class for static/class method %s" % class_name)
        attr = getattr(class_type, name)
        override_mapping = self._overrides.get(class_type)
        if override_mapping:
            override_func = override_mapping.get(name)
            if override_func:
                if is_static:
                    return override_func(attr, *args, **kwargs)
                else:
                    return override_func(class_type, attr, *args, **kwargs)
        return attr(*args, **kwargs)

    def _handle_getval(self, target, name):
        if name in self._known_vals:
            return self._known_vals[name]
        else:
            raise ValueError("Unknown value %s" % name)

    def _handle_setval(self, target, name, value):
        if name in self._known_vals:
            self._known_vals[name] = value

    def _handle_init(self, target, class_name, *args, **kwargs):
        class_type = self._known_classes.get(class_name)
        if class_type is None:
            raise ValueError("Unknown class %s" % class_name)
        # Check if __init__ is overridden
        override_mapping = self._overrides.get(class_type)
        if override_mapping:
            override_func = override_mapping.get("__init__")
            if override_func:
                return override_func(None, class_type, *args, **kwargs)
        return class_type(*args, **kwargs)

    def _handle_subclasscheck(self, target, class_name, otherclass_name, reverse=False):
        class_type = self._known_classes.get(class_name)
        if class_type is None:
            raise ValueError("Unknown class %s" % class_name)
        try:
            sub_module, sub_name = otherclass_name.rsplit(".", 1)
            __import__(sub_module, None, None, "*")
        except Exception:
            sub_module = None
        if sub_module is None:
            return False
        if reverse:
            return issubclass(class_type, getattr(sys.modules[sub_module], sub_name))
        return issubclass(getattr(sys.modules[sub_module], sub_name), class_type)

    def _handle_getclassattr(self, target, class_name, attr_name):
        # Handle class-level attribute access like EnumClass.MEMBER
        class_type = self._known_classes.get(class_name)
        if class_type is None:
            class_type = self._proxied_types.get(class_name)
        if class_type is None:
            raise ValueError("Unknown class %s" % class_name)
        return getattr(class_type, attr_name)


if __name__ == "__main__":
    max_pickle_version = int(sys.argv[1])
    config_dir = sys.argv[2]
    socket_path = sys.argv[3]
    s = Server(config_dir, max_pickle_version)
    s.serve(path=socket_path)



import os
import traceback
from .card import MetaflowCard, MetaflowCardComponent
from metaflow.extension_support import get_modules, EXT_PKG, _ext_debug

_CARD_MODULES = []


def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    import pkgutil

    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def _get_external_card_packages():
    """
    Safely extract all external card modules
    Returns:
        `list` of `ModuleType`
    """
    import importlib

    # Caching card related modules.
    global _CARD_MODULES
    if len(_CARD_MODULES) > 0:
        return _CARD_MODULES
    for m in get_modules("plugins.cards"):
        card_packages = []
        # condition checks if it is not a namespace package or is a regular package.
        if getattr(m.module, "__file__", None):
            # This supports the following cases
            # - a namespace package support with mfextinit_X.py
            # - a regular package support
            card_packages.append(m.module)
        else:
            # This is to support current system where you have a namespace package and then sub packages
            # that have __init__.py
            for _, card_mod, ispkg_c in iter_namespace(m.module):
                # Iterate submodules of metaflow_extensions.X.plugins.cards
                # For example metaflow_extensions.X.plugins.cards.monitoring
                try:
                    if not ispkg_c:
                        continue
                    cm = importlib.import_module(card_mod)
                    _ext_debug("Importing card package %s" % card_mod)
                    card_packages.append(cm)
                except Exception as e:
                    _ext_debug(
                        "External Card Module Import Exception \n\n %s \n\n %s"
                        % (str(e), traceback.format_exc())
                    )

        _CARD_MODULES.extend(card_packages)
    return _CARD_MODULES


def _load_external_cards():
    # Load external card packages
    card_packages = _get_external_card_packages()
    if not card_packages:
        return []
    external_cards = {}
    card_arr = []
    # Load cards from all external packages.
    for package in card_packages:
        try:
            cards = package.CARDS
            # Ensure that types match.
            if not type(cards) == list:
                continue
        except AttributeError as e:
            _ext_debug("Card import failed with error : %s" % str(e))
            continue
        else:
            for c in cards:
                if not isinstance(c, type) or not issubclass(c, MetaflowCard):
                    # every card should only be inheriting a MetaflowCard
                    continue
                if not getattr(c, "type", None):
                    # todo Warn user of nonexistent `type` in MetaflowCard
                    continue
                if c.type in external_cards:
                    # todo Warn user of duplicate card
                    continue
                # external_cards[c.type] = c
                _ext_debug("Adding card of type: %s" % str(c.type))
                card_arr.append(c)
    return card_arr


MF_EXTERNAL_CARDS = _load_external_cards()



from metaflow.plugins.metadata_providers.local import LocalMetadataProvider
from metaflow.metaflow_config import DATASTORE_SPIN_LOCAL_DIR


class SpinMetadataProvider(LocalMetadataProvider):
    TYPE = "spin"
    DATASTORE_DIR = DATASTORE_SPIN_LOCAL_DIR  # ".metaflow_spin"

    @classmethod
    def _get_storage_class(cls):
        from metaflow.plugins.datastores.spin_storage import SpinStorage

        return SpinStorage

    def version(self):
        return "spin"



import functools
import os
import sys

# HACK to pretend that we installed test_lib
sys.path.append(
    os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "test_lib_impl"))
)

import test_lib as lib

EXPORTED_CLASSES = {
    ("test_lib", "test_lib.alias"): {
        "TestClass1": lib.TestClass1,
        "TestClass2": lib.TestClass2,
        "BaseClass": lib.BaseClass,
        "ChildClass": lib.ChildClass,
        "ExceptionAndClass": lib.ExceptionAndClass,
        "ExceptionAndClassChild": lib.ExceptionAndClassChild,
        "TestIntEnum": lib.TestIntEnum,
        "TestStrEnum": lib.TestStrEnum,
    }
}

EXPORTED_EXCEPTIONS = {
    ("test_lib", "test_lib.alias"): {
        "SomeException": lib.SomeException,
        "MyBaseException": lib.MyBaseException,
        "ExceptionAndClass": lib.ExceptionAndClass,
        "ExceptionAndClassChild": lib.ExceptionAndClassChild,
    }
}

PROXIED_CLASSES = [functools.partial]

EXPORTED_FUNCTIONS = {"test_lib": {"test_func": lib.test_func}}

EXPORTED_VALUES = {"test_lib": {"test_value": lib.test_value}}



import platform
import sys

if sys.version_info < (3, 6):

    class Tags:
        __getattr__ = lambda self, name: (_ for _ in ()).throw(
            Exception("packaging.tags is not avaliable for Python < 3.6")
        )

    tags = Tags()
    parse_wheel_filename = lambda: (_ for _ in ()).throw(
        RuntimeError("packaging.utils is not avaliable for Python < 3.6")
    )
else:
    from metaflow._vendor.packaging import tags
    from metaflow._vendor.packaging.utils import parse_wheel_filename

from urllib.parse import unquote

from metaflow.exception import MetaflowException

MICROMAMBA_URL = "https://micro.mamba.pm/api/micromamba/{platform}/{version}"
MICROMAMBA_MIRROR_URL = "https://micromamba.outerbounds.sh/{platform}/{version}.tar.bz2"


def conda_platform():
    # Returns the conda platform for the Python interpreter
    _32_bit_interpreter = sys.maxsize <= 2**32
    if platform.system() == "Linux":
        if _32_bit_interpreter:
            return "linux-32"
        else:
            return "linux-64"
    elif platform.system() == "Darwin":
        if platform.machine() == "arm64":
            return "osx-arm64"
        elif _32_bit_interpreter:
            return "osx-32"
        else:
            return "osx-64"


def markers_from_platform(platform):
    plat, mach = platform.split("-")

    platform_system = {"osx": "Darwin", "linux": "Linux"}.get(plat)
    platform_machine = {
        "32": "x86",
        "64": "x86_64",
        "arm64": "aarch64",
        "aarch64": "aarch64",
    }.get(mach)

    markers = {
        k: v
        for k, v in {
            "platform_system": platform_system,
            "platform_machine": platform_machine,
        }.items()
        if v is not None
    }
    return markers


def wheel_tags(wheel):
    _, _, _, tags = parse_wheel_filename(wheel)
    return list(tags)


def pip_tags(python_version, mamba_platform, freetheaded=False):
    # Returns a list of pip tags containing (implementation, platforms, abis) tuples
    # assuming a CPython implementation for Python interpreter.

    # Inspired by https://github.com/pypa/pip/blob/0442875a68f19b0118b0b88c747bdaf6b24853ba/src/pip/_internal/utils/compatibility_tags.py
    py_version = tuple(map(int, python_version.split(".")[:2]))
    glibc_tags = [
        "_2_17",
        "_2_18",
        "_2_19",
        "_2_20",
        "_2_21",
        "_2_23",
        "_2_24",
        "_2_25",
        "_2_26",
        "_2_27",
        "_2_28",
        "_2_29",
        "_2_30",
        "_2_31",
        "_2_32",
        "_2_33",
        "_2_34",
        "_2_35",
        "_2_36",
        "_2_37",
        "_2_38",
    ]
    if mamba_platform == "linux-64":
        platforms = [
            "manylinux%s_x86_64" % s
            for s in [
                "1",
                "2010",
                "2014",
            ]
            + glibc_tags
        ]
        platforms.append("linux_x86_64")
    elif mamba_platform == "linux-aarch64":
        platforms = [
            "manylinux%s_aarch64" % s
            for s in [
                "2014",
            ]
            + glibc_tags
        ]
        platforms.append("linux_aarch64")
    elif mamba_platform == "osx-64":
        platforms = tags.mac_platforms(arch="x86_64")
    elif mamba_platform == "osx-arm64":
        platforms = tags.mac_platforms(arch="arm64")
    else:
        raise MetaflowException("Unsupported platform: %s" % mamba_platform)

    interpreter = "cp%s" % ("".join(map(str, py_version)))

    abis = tags._cpython_abis(py_version)
    if freetheaded:
        abis = [f"{abi}t" for abi in abis]

    supported = []
    supported.extend(tags.cpython_tags(py_version, abis, platforms))
    supported.extend(tags.compatible_tags(py_version, interpreter, platforms))
    return supported


def parse_filename_from_url(url):
    # Separate method as it might require additional checks for the parsing.
    filename = url.split("/")[-1]
    return unquote(filename)



from .inputs import Inputs
from .flow_datastore import FlowDataStore
from .datastore_set import TaskDataStoreSet
from .task_datastore import TaskDataStore
from .spin_datastore import SpinTaskDatastore
